# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch
import numpy as np
import functools
import networkx as nx
import matplotlib.pyplot as plt

from copy import deepcopy
from typing import Dict, Tuple, Union, List, Callable, Any
from tqdm import tqdm
from functools import partial

from embodichain.utils.math import pose_inv
from embodichain.utils.logger import log_info, log_warning, log_error
from embodichain.lab.sim.cfg import MarkerCfg
from embodichain.lab.gym.utils.misc import resolve_env_params, data_key_to_control_part
from embodichain.data.enum import Hints, EefExecute
from .utils import generate_affordance_from_src


# https://stackoverflow.com/questions/41834530/how-to-make-python-decorators-work-like-a-tag-to-make-function-calls-by-tag
class TagDecorator(object):
    def __init__(self, tagName):
        self.functions = {}
        self.tagName = tagName

    def __str__(self):
        return "<TagDecorator {tagName}>".format(tagName=self.tagName)

    def __call__(self, f):
        # class_key = f"{f.__module__}.{f.__qualname__.rsplit('.', 1)[0]}"
        class_name = f.__qualname__.split(".")[0]
        if class_name in self.functions.keys():
            self.functions[class_name].update({f.__name__: f})
        else:
            self.functions.update({class_name: {f.__name__: f}})
        return f


@functools.lru_cache(maxsize=None)  # memoization
def get_func_tag(tagName):
    return TagDecorator(tagName)


tag_node = get_func_tag("node")
tag_edge = get_func_tag("edge")


class ActionBank:
    _function_type: Dict[str, Callable]

    def __init__(self, conf: Dict):
        self.conf = conf

    @property
    def vis_gantt(self):
        return self.conf.get("misc", {}).get("vis_gantt", False)

    @property
    def vis_graph(self):
        return self.conf.get("misc", {}).get("vis_graph", False)

    @property
    def warpping(self):
        return self.conf.get("misc", {}).get("warpping", True)

    @staticmethod
    def get_function_name(input: Dict) -> str:
        """
        Retrieve the function name from the input dictionary.

        This method assumes that the input dictionary contains exactly one key,
        which represents the function name. If the dictionary contains more than
        one key, a ValueError is raised.

        Args:
            input (Dict): A dictionary with a single key representing the function name.

        Returns:
            str: The function name extracted from the dictionary.

        Raises:
            ValueError: If the input dictionary contains zero or more than one key.
        """
        if len(list(input.keys())) != 1:
            raise ValueError(
                "The input dict {} has invalid keys {}.".format(
                    input, list(input.keys())
                )
            )

        return list(input.keys())[0]

    def get_scope_names(
        self,
    ) -> List[str]:
        return list(self.conf["scope"].keys())

    def get_node_names(self, bool_attr_name: str = None) -> Dict[str, List[str]]:
        scopes = self.get_scope_names()
        nodes = self.conf["node"]
        node_names = {}
        for scope in scopes:
            node_names[scope] = []
            for node in nodes[scope]:
                if bool_attr_name is not None:
                    if node[self.get_function_name(node)].get(bool_attr_name, False):
                        function_name = ActionBank.get_function_name(node)
                        node_names[scope].append(function_name)
        return node_names

    def graph2id(self, type: str = "node") -> Dict[str, Dict[str, str]]:
        scopes = self.get_scope_names()
        nodes = self.conf[type]
        graph_2_id = {}
        for scope in scopes:
            graph_2_id[scope] = {}
            for i, node in enumerate(nodes[scope]):
                function_name = ActionBank.get_function_name(node)
                graph_2_id[scope].update({function_name: i})
        return graph_2_id

    def get_edge_names(self, node_name: str = None) -> Dict[str, List[Dict[str, str]]]:
        scopes = self.get_scope_names()
        edges = self.conf["edge"]
        edge_names = {}
        for i, key in enumerate(scopes):
            edge_names[key] = []
            for edge in edges[key]:
                function_name = ActionBank.get_function_name(edge)
                src = edge[function_name]["src"]
                sink = edge[function_name]["sink"]
                temp = {"name": function_name, "src": src, "sink": sink}
                edge_names[key].append(temp)
        if node_name is None:
            return edge_names
        else:
            filtered_edge_names = {}
            for scope, edge_list in edge_names.items():
                filtered_edge_names[scope] = [
                    edge
                    for edge in edge_list
                    if edge["src"] == node_name or edge["sink"] == node_name
                ]
            return filtered_edge_names

    def _infer_fill_type(self, scope: str, label: str, edge_cfg: Dict) -> str:
        # 1) explicit in config
        ft = edge_cfg.get("fill_type", None)
        # 2) built-in eef rules
        fn = edge_cfg.get("name", None)
        if fn in {EefExecute.OPEN.value, EefExecute.CLOSE.value}:
            return "still"
        # 3) explicit wins; otherwise default
        return ft if ft in ("still", "scalable") else "still"

    def _get_unit_pairs(self, legends: List[str]) -> Dict[str, str]:
        """
        Return a symmetric map executor -> partner within the same unit (arm+eef).
        Priority:
          1) explicit config: self.conf["misc"]["unit_pairs"] = [["right_arm","right_eefhand"], ["left_arm","left_eefhand"], ...]
          2) heuristic by side-prefix and name hints ('arm' vs 'eef'/'hand'/'gripper')
        """
        pairs: Dict[str, str] = {}

        # 1) explicit mapping if provided
        explicit = self.conf.get("misc", {}).get("unit_pairs", None)
        if explicit:
            for a, b in explicit:
                if a in legends and b in legends:
                    pairs[a] = b
                    pairs[b] = a

        # 2) heuristic fallback
        def side_key(name: str) -> str:
            # prefer token before '_' or '-', else prefix match
            if "_" in name:
                return name.split("_", 1)[0]
            if "-" in name:
                return name.split("-", 1)[0]
            for pref in ("left", "right", "L", "R"):
                if name.lower().startswith(pref.lower()):
                    return pref
            return ""

        eef_hints = Hints.EEF.value
        arm_hints = Hints.ARM.value

        from collections import defaultdict

        by_side = defaultdict(list)
        for n in legends:
            by_side[side_key(n)].append(n)

        for _, names in by_side.items():
            arms = [n for n in names if any(h in n.lower() for h in arm_hints)]
            eefs = [n for n in names if any(h in n.lower() for h in eef_hints)]
            # pair the first unmatched arm with the first unmatched eef
            for a in arms:
                if a in pairs:
                    continue
                partner = next((e for e in eefs if e not in pairs), None)
                if partner:
                    pairs[a] = partner
                    pairs[partner] = a

        return pairs

    def _apply_bubble_filling(self, packages, taskkey2index):
        from collections import defaultdict

        if not packages:
            return packages

        # group by executor
        per_legend = defaultdict(list)
        for p in packages:
            per_legend[p["legend"]].append(p)
        for lg in per_legend:
            per_legend[lg].sort(key=lambda x: (x["start"], x["end"]))

        legends = list(per_legend.keys())
        unit_pairs = self._get_unit_pairs(legends)

        # fill_type lookup
        fill_type = {}
        for scope, scope_edges in self.conf.get("edge", {}).items():
            for edge in scope_edges:
                lbl = list(edge.keys())[0]
                fill_type[lbl] = edge[lbl].get("fill_type", "still")

        label2pkg = {p["label"]: p for p in packages}
        global_end = max(p["end"] for p in packages)

        def first_start_at_or_after(seq, t):
            for pkg in seq:
                if pkg["start"] >= t:
                    return pkg["start"]
            return global_end

        # optional sync boundary (unchanged)
        dep_of = defaultdict(list)
        for e_label, s in self.conf.get("sync", {}).items():
            for d in s.get("depend_tasks", []):
                dep_of[d].append(e_label)

        def sync_boundary_for(lbl):
            deps = dep_of.get(lbl, [])
            if not deps:
                return None
            starts = [label2pkg[d]["start"] for d in deps if d in label2pkg]
            return max(starts) if starts else None

        # unit-aware filling
        for lg, seq in per_legend.items():
            partner = unit_pairs.get(lg, None)
            partner_seq = per_legend.get(partner, []) if partner else []

            # middle gaps
            for i in range(len(seq) - 1):
                curr, nxt = seq[i], seq[i + 1]
                if curr["end"] < nxt["start"]:
                    cap_local = nxt["start"]
                    cap_partner = (
                        first_start_at_or_after(partner_seq, curr["end"])
                        if partner_seq
                        else global_end
                    )
                    cap_sync = sync_boundary_for(curr["label"])
                    cap = (
                        min(cap_local, cap_partner, cap_sync)
                        if cap_sync is not None
                        else min(cap_local, cap_partner)
                    )
                    if cap > curr["end"]:
                        curr["end"] = cap
                        curr.setdefault(
                            "fill_type", fill_type.get(curr["label"], "still")
                        )

            # tail gap: cap by partner’s next (≥ end), not global
            if seq:
                last = seq[-1]
                cap_partner = (
                    first_start_at_or_after(partner_seq, last["end"])
                    if partner_seq
                    else last["end"]
                )
                if cap_partner > last["end"]:
                    last["end"] = cap_partner
                    last.setdefault("fill_type", fill_type.get(last["label"], "still"))

        return packages

    def parse_network(
        self,
        node_functions: Dict[str, Callable],
        edge_functions: Dict[str, Callable],
        vis_graph: bool = False,
    ) -> Tuple[nx.DiGraph, Dict[str, List], Dict[str, Tuple[int, int]]]:
        """Construct a graph with self.conf["node"]&["edge"], and node_functions, edge_functions be its node generator and edge linker.

        Return the constructed nx.DiGraph graph_compose,

        and tasks_data = {"scope name" : [(scope_id=task_id, skill_duration_{i})]},

        and taskkey2index = {"edge_name": (scope_id=task_id, edge_id=skill_id)}

        Args:
            node_functions (Dict[str, Callable]): A Dict consists of key-value pair that key be all nodes (affordance) name and value be its generating functions
            edge_functions (Dict[str, Callable]): A Dict consists of key-value pair that key be all edges (skill, a part of a trajectory) name and value be its linker functions
            vis_graph (bool, optional): Whether to show the graph or not. Defaults to True.

        Returns:
            graph_compose: A composed nx.DiGraph representing the graph defined in self.conf, while the node generators and edge linkers prepared,
            tasks_data: A Dict consists of key-value pair that key be all scopes' names, and value be List(Tuple=(scope_id=task_id, skill_duration))
            taskkey2index: A Dict consists of key-value pair that key be all edges' names, and value be Tuple=(scope_id=task_id, edge_id=skill_id)
        """
        nodes = self.conf.get("node", {})
        edges = self.conf.get("edge", {})
        graph_type = self.conf.get("scope", {})

        graphs = {key: nx.DiGraph() for key in graph_type.keys()}
        disjoint_names = {}
        tasks_data = {}
        taskkey2index = {}

        # key2index = {}
        edges_flatten = {}
        for i, key in enumerate(graphs.keys()):
            # key2index[key] = i
            for j, edge in enumerate(edges[key]):
                edge = deepcopy(edge)
                taskkey2index[ActionBank.get_function_name(edge)] = (i, j)
                edge["type"] = key
                edges_flatten.update(edge)
        for i, key in enumerate(graphs.keys()):
            tasks_data[key] = []
            for edge in edges[key]:
                label = ActionBank.get_function_name(edge)  # edge label in config
                cfg = edge[label]
                src = cfg["src"]
                sink = cfg["sink"]
                kwargs = cfg.get("kwargs", {})
                duration = cfg.get("duration", 0)
                if not isinstance(duration, int):
                    raise TypeError("Duration must be an integer.")

                # function to call
                fn_name = cfg.get("name", label)
                # normalize and persist fill_type (default + built-in rules)
                fill_type = self._infer_fill_type(key, label, cfg)

                graphs[key].add_edge(
                    src,
                    sink,
                    linker=partial(
                        edge_functions[fn_name], **kwargs, duration=duration
                    ),
                    duration=duration,
                    fill_type=fill_type,
                    edge_label=label,
                    scope=key,
                )
                tasks_data[key].append((i, duration))

            for node in nodes[key]:
                function_name = ActionBank.get_function_name(node)
                if function_name in disjoint_names.keys():
                    error_msg = f"Function {function_name} is already defined in {disjoint_names[function_name]} but re-defined in {key} again."
                    log_error(error_msg)
                disjoint_names.update({function_name: key})
                graphs[key].add_node(
                    function_name,
                    generator=partial(
                        node_functions[node[function_name]["name"]],
                        **node[function_name]["kwargs"],
                    ),
                )

        graph_compose = nx.DiGraph()
        for key, graph in graphs.items():
            if self.vis_graph or vis_graph:
                nx.draw(graph, with_labels=True)
                plt.show()
            if graph_type[key]["type"] == "tree":
                assert nx.is_tree(
                    graph.to_undirected()
                ), "{} graph is not tree.".format(key)

            graph_compose = nx.compose(graph_compose, graph)

        if self.vis_graph or vis_graph:
            nx.draw(graph_compose, with_labels=True)
            plt.show()

        return graph_compose, tasks_data, taskkey2index

    def gantt(
        self,
        tasks_data: Dict[str, List],
        taskkey2index: Dict[str, int],
        vis: bool = False,
    ) -> Dict[str, Any]:
        """Given tasks on different machines and skills within tasks that takes a specific duration, try to minimize the max length among task, while respecting:
           Constraint 1: For skills of a same task, which occupied a same machine, do not overlap with each other.
           Constraint 2: For skills of a same task, the start time of skill should not surpass the end time of the before skill
           Constraint 3: For sync edges define in self.conf["sync"], which defined a skill, its start time should not surpass the end time of the depend skill.
           with a set of start and end time of all skills. Then draw the gantt with the solution, return the solution packages with start and end time of each edge.

        Args:
            tasks_data (Dict[str, List]): A Dict consists of key-value pair that key be all scopes' names, and value be List(Tuple=(scope_id=task_id, skill_duration))
            taskkey2index (Dict[str, int]): A Dict consists of key-value pair that key be all edges' names, and value be Tuple=(scope_id=task_id, edge_id=skill_id)
            vis (bool, optional): Whether to visualize the gantt or not. Defaults to False.

        Returns:
            packages Dict[str, Any]:
                {
                    "total_num": int(solver.objective_value)=max length of a task (among tasks = scopes = machines = executors),
                    "packages": packages=List(Dict=assigned_task={
                        "labels": edge_name,
                        "start" : edge.start
                        "end": edge.start + edge.duration
                        "legend": task_name
                        "color": color representing the task_id, the former the bluer
                    }
                )
            }
        """
        import collections

        # https://developers.google.com/optimization/scheduling/task_shop?hl=zh-cn
        from ortools.sat.python import cp_model

        machines_count = len(list(tasks_data.keys()))
        all_machines = range(machines_count)
        # Computes horizon dynamically as the sum of all durations.
        id2key = {i: key for i, key in enumerate(tasks_data.keys())}
        tasks_data = list(tasks_data.values())

        horizon = sum(skill[1] for task in tasks_data for skill in task)
        model = cp_model.CpModel()
        # Named tuple to store information about created variables.
        skill_type = collections.namedtuple("skill_type", "start end interval")

        # Creates task intervals and add to the corresponding machine lists.
        all_skills = {}
        machine_to_intervals = collections.defaultdict(list)

        for task_id, task in enumerate(tasks_data):
            for skill_id, skill in enumerate(task):
                machine, duration = skill
                suffix = f"_{task_id}_{skill_id}"
                start_var = model.new_int_var(0, horizon, "start" + suffix)
                end_var = model.new_int_var(0, horizon, "end" + suffix)
                interval_var = model.new_interval_var(
                    start_var, duration, end_var, "interval" + suffix
                )
                all_skills[task_id, skill_id] = skill_type(
                    start=start_var, end=end_var, interval=interval_var
                )
                machine_to_intervals[machine].append(interval_var)

        # Create and add disjunctive constraints for each machine.
        for machine in all_machines:
            model.add_no_overlap(machine_to_intervals[machine])

        # Precedences inside a task.
        for task_id, task in enumerate(tasks_data):
            for skill_id in range(len(task) - 1):
                model.add(
                    all_skills[task_id, skill_id + 1].start
                    >= all_skills[task_id, skill_id].end
                )

        sync_edges = self.conf["sync"]
        for edge_name in sync_edges.keys():
            task_id, skill_id = taskkey2index[edge_name]
            for depend_task in sync_edges[edge_name]["depend_tasks"]:
                before_task_id, before_skill_id = taskkey2index[depend_task]
                model.add(
                    all_skills[task_id, skill_id].start
                    >= all_skills[before_task_id, before_skill_id].end
                )

        # Makespan objective.
        obj_var = model.new_int_var(0, horizon, "makespan")

        max_equality = []
        for task_id, task in enumerate(tasks_data):
            if len(task) != 0:
                max_equality.append(all_skills[task_id, len(task) - 1].end)

        model.add_max_equality(obj_var, max_equality)
        model.minimize(obj_var)
        solver = cp_model.CpSolver()
        status = solver.solve(model)

        packages = []
        if status == cp_model.OPTIMAL or status == cp_model.FEASIBLE:
            # Create one list of assigned skills per machine.
            n = len(list(id2key.keys()))
            color = [(1 - i / n, 0, i / n) for i in range(n)]
            keys_ = list(taskkey2index.keys())
            values_ = list(taskkey2index.values())
            for task_id, task in enumerate(tasks_data):
                for skill_id, skill in enumerate(task):
                    machine = skill[0]
                    duration = skill[1]
                    start = solver.value(all_skills[task_id, skill_id].start)

                    assigned_task = {}
                    assigned_task["label"] = keys_[values_.index((task_id, skill_id))]
                    assigned_task["start"] = start
                    assigned_task["end"] = start + duration
                    assigned_task["legend"] = id2key[task_id]
                    assigned_task["color"] = color[task_id]
                    packages.append(assigned_task)

            # Finally print the solution found.
            log_warning(f"Optimal Schedule Length: {solver.objective_value}")
        else:
            log_error("No solution found.")

        packages = self._apply_bubble_filling(packages, taskkey2index)

        new_total = max((p["end"] for p in packages), default=0)

        if self.vis_gantt or vis:
            from embodichain.utils.visualizer import Gantt

            draw_gantt_data = {
                "title": " Sample GANTT",
                "xlabel": "Trajectory (steps)",
                "packages": packages,
            }
            g = Gantt(draw_gantt_data)
            g.render()
            g.show()

        return {"total_num": int(new_total), "packages": packages}

    def initialize_action_list(
        self, env, action_list, executor: str, executor_init_info: Dict
    ) -> np.ndarray:
        """
        Initialize the action list for a specific executor.

        This method initializes the action trajectory for the given executor based on the provided initialization information.
        The initialization can be done using predefined qpos values or the current qpos of the executor.

        Args:
            self (ActionBank): The ActionBank instance.
            env (object): The environment instance containing executor and affordance data.
            action_list (np.ndarray): A numpy array of shape (T, qpos_dim), representing the uninitialized action trajectory for the executor.
            executor (str): The name of the executor (e.g., "left_arm", "right_arm").
            executor_init_info (Dict): A dictionary containing initialization information for the executor, such as method and parameters.

        Returns:
            np.ndarray: The initialized action list for the executor.
        """

        def initialize_with_given_qpos(action_list, executor, executor_init_info, env):
            given_qpos = executor_init_info.get("kwargs", {}).get("given_qpos", None)
            if given_qpos is None:
                log_warning(
                    "No given_qpos is provided for initialize_with_given_qpos. Using {}.".format(
                        "{}_{}_qpos".format(executor, "init")
                    )
                )
                given_qpos = env.affordance_datas["{}_{}_qpos".format(executor, "init")]

            executor_qpos_dim = action_list[executor].shape[0]
            given_qpos = np.asarray(given_qpos)
            if len(given_qpos.shape) != 1:
                log_warning(
                    f"Shape of given init qpos should be (1,), but got {given_qpos.shape} with length {len(given_qpos.shape)}. Using 0-th element with {given_qpos.shape[-1]}."
                )
                last_ids = (0,) * (given_qpos.ndim - 1) + (Ellipsis,)
                given_qpos = given_qpos[last_ids]

            if given_qpos.shape[0] != executor_qpos_dim:
                log_error(
                    f"Shape of given init qpos should be {(executor_qpos_dim,)}, but got {given_qpos.shape[0]}."
                )

            init_node_name = executor_init_info.get(
                "init_node_name", f"{executor}_init_qpos"
            )
            if (
                len(init_node_name) > 0
            ):  # so if you don't need to inject it, just assign the "init_node_name" to be "" in action_config
                env.affordance_datas[init_node_name] = given_qpos
            action_list[executor][:, 0] = given_qpos

            return action_list

        def initialize_with_current_qpos(
            action_list, executor, executor_init_info, env
        ):
            # TODO: Hard to get current qpos for multi-agent env
            current_qpos = env.robot.get_qpos()
            joint_ids = env.robot.get_joint_ids(name=get_control_part(env, executor))

            # Handle multi-environment case
            if current_qpos.ndim == 2:
                # current_qpos shape: [num_envs, num_joints]
                # Take first environment and then select joints
                current_qpos = current_qpos[0, joint_ids].cpu()
            else:
                # Single environment case
                # current_qpos shape: [num_joints]
                current_qpos = current_qpos[joint_ids].cpu()

            executor_qpos_dim = action_list[executor].shape[0]

            # NOTE: hard code!
            current_qpos = current_qpos[:executor_qpos_dim]

            if current_qpos.shape[0] != executor_qpos_dim:
                log_error(
                    f"Shape of given init qpos should be {(executor_qpos_dim,)}, but got {current_qpos.shape[0]}."
                )

            init_node_name = executor_init_info.get(
                "init_node_name", f"{executor}_init_qpos"
            )
            if (
                len(init_node_name) > 0
            ):  # so if you don't need to inject it, just assign the "init_node_name" to be "" in action_config
                env.affordance_datas[init_node_name] = current_qpos
            action_list[executor][:, 0] = current_qpos

            return action_list

        INIT_METHOD_MAPPING = {
            "given_qpos": initialize_with_given_qpos,
            "current_qpos": initialize_with_current_qpos,
        }

        init_method = executor_init_info.get("method", "current_qpos")

        if init_method is None:
            log_warning(
                f"No init method provided in action config for executor {executor}, please check. Skipping.."
            )
            return action_list
        if init_method not in INIT_METHOD_MAPPING:
            log_warning(
                f"Provided init method action config for executor {executor}: {init_method} is not accomplished yet, please check. Skipping.."
            )
            return action_list
        else:
            init_func = INIT_METHOD_MAPPING[init_method]
            action_list = init_func(action_list, executor, executor_init_info, env)

        return action_list

    @staticmethod
    @tag_node
    @resolve_env_params
    def generate_affordances_from_src(env, affordance_infos: List[Dict]) -> bool:
        for affordance_info in affordance_infos:
            src_key = affordance_info["src_key"]
            dst_key = affordance_info["dst_key"]
            valid_funcs_name_kwargs_proc = affordance_info[
                "valid_funcs_name_kwargs_proc"
            ]
            to_array = env.action_bank.warpping

            ret = generate_affordance_from_src(
                env, src_key, dst_key, valid_funcs_name_kwargs_proc, to_array
            )
            if not ret:
                return False
        return True

    def _prepare_warpping(self, env):
        if hasattr(env, "affordance_datas"):
            for affordance_name, affordance_value in env.affordance_datas.items():
                # NOTE: take only first arena's affordance data
                if affordance_value.ndim == 3:
                    affordance_value = affordance_value[0]
                if isinstance(affordance_value, torch.Tensor):
                    affordance_value = np.asarray(affordance_value.cpu())
                env.affordance_datas[affordance_name] = affordance_value
        else:
            log_warning("No env.affordance_datas, skip _prepare_warpping..")

    def create_action_list(
        self, env, graph_compose: nx.DiGraph, packages: List[Dict], **kwargs
    ) -> Dict:
        """Create an action list based on the given environment, graph, and packages.

        Args:
            env (embodichain.lab.gym.envs.BaseEnv): The environment instance.
            graph_compose (nx.DiGraph): The composed graph containing nodes and edges.
            packages (List[Dict]): The task packages with scheduling information.

        Returns:
            Dict: The generated action list for all executors.
        """

        def initialize_action_list(
            scope: Dict, total_num: int
        ) -> Tuple[Dict, Dict, Dict]:
            """Initialize action list and related variables."""
            action_list = {}
            end_time = {}
            in_working = {}

            for executor in scope.keys():
                end_time[executor] = 0
                in_working[executor] = False

                action_list[executor] = np.zeros(
                    tuple(scope[executor]["dim"]) + (total_num,),
                    dtype=getattr(np, scope[executor]["dtype"]),
                )

                init_info = scope[executor].get("init", {})
                action_list = self.initialize_action_list(
                    env, action_list, executor, init_info
                )

            return action_list, end_time, in_working

        def generate_nodes(graph_compose: nx.DiGraph, nodes: Dict) -> bool:
            """Generate nodes using the graph's node generators."""
            node_generators = nx.get_node_attributes(graph_compose, "generator")

            failed_nodes = []
            log_info("Action bank start node generation for action graph...")
            for node_dict_list in nodes.values():
                for node in node_dict_list:
                    node_name = list(node.keys())[0]
                    try:
                        log_info(f"\tGenerating node '{node_name}' .")
                        ret = node_generators[node_name](env, **kwargs)
                        if not ret:
                            log_warning(f"Node '{node_name}' generation fails.")
                            failed_nodes.append(node_name)
                    except KeyError as e:
                        log_warning(
                            f"[KeyError] '{node_name}': {e}. Node generator might be missing or invalid."
                        )
                        failed_nodes.append(node_name)
                    except AttributeError as e:
                        log_warning(
                            f"[AttributeError] '{node_name}': {e}. Missing required attributes in environment."
                        )
                        failed_nodes.append(node_name)
                    except TypeError as e:
                        log_warning(
                            f"[TypeError] '{node_name}': {e}. Check input data types."
                        )
                        failed_nodes.append(node_name)
                    except ValueError as e:
                        log_warning(
                            f"[ValueError] '{node_name}': {e}. Check input values."
                        )
                        failed_nodes.append(node_name)
                    except Exception as e:
                        log_warning(
                            f"[UnexpectedError] '{node_name}': {e}. Debug dependencies or implementation."
                        )
                        failed_nodes.append(node_name)
            if failed_nodes:
                log_warning(f"Failed to generate the following nodes: {failed_nodes}")
                return False

            log_info(
                f"Node generation is finished. Total nodes generated: {sum(len(v) for v in nodes.values())}."
            )
            return True

        def generate_edges(
            total_num: int,
            all_executors: List[str],
            edges_flatten: Dict,
            node_linkers: Dict,
        ) -> None:
            """
            Generate edges and populate the action list for all executors.

            Args:
                total_num (int): The total number of time steps for the action list.
                all_executors (List[str]): A list of executor names (e.g., "left_arm", "right_arm").
                edges_flatten (Dict[str, Dict]): A flattened dictionary of edges, where keys are edge labels
                    and values are dictionaries containing edge details (e.g., "src", "sink").
                node_linkers (Dict[Tuple[str, str], Callable]): A dictionary mapping edge (source, sink) pairs
                    to their corresponding linker functions.

            Returns:
                None: This function modifies the `action_list` in place.
            """

            def get_task_in_time(tasks, time):
                """Get the task that is active at the given time."""
                return next(
                    (task for task in tasks if task["start"] <= time < task["end"]),
                    None,
                )

            for i in tqdm(range(total_num), desc="Generating edges"):
                for executor in all_executors:
                    if end_time[executor] == i:
                        in_working[executor] = False

                    if not in_working[executor]:
                        pkg = get_task_in_time(
                            [
                                pkg
                                for pkg in packages["packages"]
                                if pkg["legend"] == executor
                            ],
                            i,
                        )
                        if pkg is None:
                            if i >= 1:
                                action_list[executor][..., i] = action_list[executor][
                                    ..., i - 1
                                ]
                        else:
                            end_time[executor] = pkg["end"]
                            skill_idx = (
                                edges_flatten[pkg["label"]]["src"],
                                edges_flatten[pkg["label"]]["sink"],
                            )
                            ret = node_linkers[skill_idx](env)
                            if not isinstance(ret, np.ndarray):

                                if isinstance(ret, torch.Tensor):
                                    ret = ret.cpu().numpy()
                                else:
                                    raise TypeError(
                                        "The return value of the linker {} must be a numpy array, but a {}.".format(
                                            skill_idx, type(ret)
                                        )
                                    )

                            start_idx = pkg["start"]
                            end_idx = pkg["end"]

                            T_need = end_idx - start_idx
                            T_orig = ret.shape[1]

                            # fill_type of this edge
                            ft = edges_flatten[pkg["label"]].get(
                                "fill_type", pkg.get("fill_type", "still")
                            )

                            def _resample_time(x, new_T):
                                if new_T == x.shape[1]:
                                    return x
                                if x.shape[1] <= 1:
                                    return np.repeat(x, new_T, axis=1)[:, :new_T]
                                t_old = np.linspace(0.0, 1.0, x.shape[1])
                                t_new = np.linspace(0.0, 1.0, new_T)
                                out = np.empty((x.shape[0], new_T), dtype=x.dtype)
                                for d in range(x.shape[0]):
                                    out[d] = np.interp(t_new, t_old, x[d])
                                return out

                            def _pad_or_trim_last(x, new_T):
                                if new_T <= x.shape[1]:
                                    return x[:, :new_T]
                                pad = np.repeat(x[:, -1:], new_T - x.shape[1], axis=1)
                                return np.concatenate([x, pad], axis=1)

                            if T_need != T_orig:
                                if ft == "scalable":
                                    ret = _resample_time(ret, T_need)
                                else:  # "still"
                                    ret = _pad_or_trim_last(ret, T_need)

                            action_list[executor][..., start_idx:end_idx] = ret
                            in_working[executor] = True

        # Main logic
        scope = self.conf["scope"]
        total_num = packages["total_num"]
        all_executors = list(scope.keys())
        edges_flatten = {
            k: v
            for edges in self.conf["edge"].values()
            for edge in edges
            for k, v in edge.items()
        }
        node_linkers = nx.get_edge_attributes(graph_compose, "linker")

        action_list, end_time, in_working = initialize_action_list(scope, total_num)

        if self.warpping:
            self._prepare_warpping(env)

        if not generate_nodes(graph_compose, self.conf["node"]):
            return None

        # After node initialization, check if env.affordance_datas contains updated initial value for each executor.
        for executor in scope.keys():
            init_node_name = "{}_{}_qpos".format(executor, "init")
            if (
                not hasattr(env, "affordance_datas")
                or init_node_name not in env.affordance_datas
            ):
                log_warning(
                    f"Executor '{executor}': init_node_name '{init_node_name}' not found in env.affordance_datas. Skipping initial value update."
                )
                continue
            affordance_init = env.affordance_datas[init_node_name]
            affordance_init = np.asarray(affordance_init)
            action_init_slice = action_list[executor][:, 0]
            if affordance_init.shape != action_init_slice.shape:
                log_warning(
                    f"Executor '{executor}': affordance_init shape {affordance_init.shape} does not match action_list[executor][:, 0] shape {action_init_slice.shape}. Skipping initial value update."
                )
                continue
            if not np.allclose(action_init_slice, affordance_init):
                log_info(
                    f"Updated initial value for executor '{executor}' in action_list from affordance_datas['{init_node_name}']."
                )
                action_list[executor][:, 0] = affordance_init

        generate_edges(total_num, all_executors, edges_flatten, node_linkers)

        return action_list


def attach_node_and_edge(
    cls: ActionBank, functions_dict: Dict[str, Dict[str, Callable]]
) -> ActionBank:
    for tag, funcs in functions_dict.items():
        tag_function = get_func_tag(tag)
        for func_name, func in funcs.items():
            setattr(cls, func_name, staticmethod(func))

            class_name = cls.__name__
            if class_name in tag_function.functions.keys():
                tag_function.functions[class_name].update({func_name: func})
            else:
                tag_function.functions.update({class_name: {func_name: func}})
    return cls


def attach_action_bank(cls, action_bank: ActionBank, **kwargs):
    def set_attr_for_cls(cls, attr_name: str, attr_value: Any):
        if hasattr(cls, attr_name):
            getattr(cls, attr_name).append(attr_value)
        else:
            setattr(cls, attr_name, [attr_value])

    action_config = kwargs.get("action_config", None)
    if action_config is None:
        log_error(
            f"The action config is None, but it's needed for Env: {type(cls).__name__}, Task Type: {cls.metadata['task_type']}."
        )
    set_attr_for_cls(cls, "action_banks", action_bank(action_config))

    vis_graph = kwargs.get("vis", False)
    graph_compose, jobs_data, jobkey2index = cls.action_banks[-1].parse_network(
        get_func_tag("node").functions[cls.action_banks[-1].__class__.__name__],
        get_func_tag("edge").functions[cls.action_banks[-1].__class__.__name__],
        vis_graph=vis_graph,
    )

    vis_gantt = kwargs.get("vis", False)
    package = cls.action_banks[-1].gantt(jobs_data, jobkey2index, vis=vis_gantt)

    set_attr_for_cls(cls, "packages", package)
    set_attr_for_cls(cls, "graph_composes", graph_compose)

    return cls


def get_xpos_name(affordance_name: str) -> str:
    if affordance_name.find("qpos") == -1:
        affordance_xpos_name = affordance_name + "_xpos"
    else:
        affordance_xpos_name = affordance_name.replace("qpos", "xpos")
    return affordance_xpos_name


def get_control_part(env, agent_uid):
    control_parts = env.metadata["dataset"]["robot_meta"].get("control_parts", [])

    if agent_uid in control_parts:
        return agent_uid
    else:
        return data_key_to_control_part(
            robot=env.robot,
            control_parts=control_parts,
            data_key=agent_uid,
        )


def generate_trajectory_qpos(
    env,
    agent_uid: str,
    trajectory: Dict[str, np.ndarray],
    trajectory_id: str,
    gather_index: List[int],
    trajectory_index: int,
    affordance_name: str,
    slaver: str = "",
    canonical_trajectory: List[float] = None,
    canonical_trajectory_index: int = None,
    canonical_pose: List[float] = [],
    vis: bool = False,
) -> bool:
    affordance_xpos_name = get_xpos_name(affordance_name)

    current_qpos = torch.as_tensor(trajectory[trajectory_id])[trajectory_index][
        None, gather_index
    ]  # TODO: only for 1 env
    control_part = get_control_part(env, agent_uid)
    try:
        affordance_xpos = env.robot.compute_fk(
            torch.as_tensor(current_qpos),
            control_part,
            to_matrix=True,
        )
    except RuntimeError as e:
        log_warning(f"control part {control_part} has no solver.")
        affordance_xpos = torch.zeros((1, 4, 4), device=current_qpos.device)
    if slaver != "":
        assert canonical_trajectory is not None
        assert canonical_trajectory_index is not None
        assert (
            len(canonical_pose) == 4
        ), f"canonical_pose should be a 4x4 matrix, but got {len(canonical_pose)} elements."
        canonical_pose = torch.as_tensor(
            canonical_pose,
            device=affordance_xpos.device,
            dtype=affordance_xpos.dtype,
        ).reshape(1, 4, 4)
        can_affordance_xpos = env.robot.compute_fk(
            torch.as_tensor(canonical_trajectory)[canonical_trajectory_index][
                gather_index
            ],
            get_control_part(env, agent_uid),
            to_matrix=True,
        )
        can_obj_xpos = canonical_pose
        obj_xpos = env.sim.get_asset(slaver).get_local_pose(to_matrix=True)
        affordance_xpos = torch.bmm(
            obj_xpos, torch.bmm(pose_inv(can_obj_xpos), can_affordance_xpos)
        )
        control_part = get_control_part(env, agent_uid)
        qpos_seed = env.robot.get_qpos()[:, env.robot.get_joint_ids(name=control_part)]
        ret, current_qpos = env.robot.compute_ik(
            affordance_xpos, qpos_seed, control_part
        )
        ret = ret.all().item()
        if not ret:
            log_warning(
                f"IK failed for slaver {slaver} with xpos {affordance_xpos}. Using the previous qpos instead."
            )
            return False

    if vis:
        env.sim.draw_marker(
            cfg=MarkerCfg(
                marker_type="axis",
                axis_xpos=affordance_xpos,
                axis_size=0.002,
                axis_len=0.005,
            )
        )
    # TODO: only support 1 env numpy now
    current_qpos = current_qpos.squeeze(0).cpu().numpy()
    affordance_xpos = affordance_xpos.squeeze(0).cpu().numpy()

    env.affordance_datas[affordance_name] = current_qpos
    env.affordance_datas[affordance_xpos_name] = affordance_xpos
    return True


def modify_action_config_edges(
    action_config: Dict,
    duration_updates: Dict[str, int] = None,
    trajectory_updates: Dict[str, List] = None,
    analytic_planner: bool = False,
) -> Dict:
    """
    Modify the action configuration by updating the duration and trajectory of edges.

    This function iterates through all edges in the action configuration and applies updates to their
    duration and trajectory based on the provided mappings. If `analytic_planner` is enabled, the edge
    name is set to "plan_trajectory".

    Args:
        action_config (Dict): The original action configuration.
        duration_updates (Dict[str, int], optional): A mapping of edge names to their new durations.
        trajectory_updates (Dict[str, List], optional): A mapping of edge names to their new trajectories.
        analytic_planner (bool, optional): If True, sets the edge name to "plan_trajectory". Defaults to False.

    Returns:
        Dict: The modified action configuration.
    """
    modified_config = deepcopy(action_config)

    # Iterate through all scopes in the action configuration
    for scope_name, scope_edges in modified_config["edge"].items():
        for edge_config in scope_edges:
            edge_name = list(edge_config.keys())[0]
            edge_data = edge_config[edge_name]
            # If analytic_planner is enabled, set the edge name to "plan_trajectory"
            if analytic_planner:
                edge_data["name"] = "plan_trajectory"

            # Update the duration if a mapping is provided
            if duration_updates and edge_name in duration_updates:
                edge_data["duration"] = duration_updates[edge_name]

            # Update the trajectory if a mapping is provided
            if trajectory_updates and edge_name in trajectory_updates:
                edge_data.setdefault("kwargs", {})  # Ensure "kwargs" exists
                edge_data["kwargs"]["trajectory"] = trajectory_updates[edge_name]

    return modified_config


def to_affordance_name(name: str) -> str:
    return name.replace("generate_", "")


def to_affordance_node_func(name: str) -> str:
    return "generate_" + name


class GeneralActionBank(ActionBank):
    @staticmethod
    @tag_edge
    def load_trajectory(
        env,
        trajectory_id: str,
        gather_index: List[int],
        keypose_timesteps: Tuple[int, int],
        raw_duration: int,
        duration: int,
        **kwargs,
    ):
        from scipy import interpolate

        f = {}
        start_t, end_t = keypose_timesteps[0], keypose_timesteps[1]
        trajectory = np.asarray(env.trajectory[trajectory_id])[:, gather_index]
        sub_trajectory = trajectory[start_t:end_t, :]
        ds_sub_trajectory = np.zeros((duration, sub_trajectory.shape[1]))
        for i in range(sub_trajectory.shape[1]):
            x = np.arange(sub_trajectory.shape[0])
            f[i] = interpolate.interp1d(x, sub_trajectory[:, i], axis=-1)
            ds_sub_trajectory[:, i] = f[i](np.linspace(0, raw_duration - 1, duration))

        return ds_sub_trajectory.T  # (D, T)

    @staticmethod
    @tag_edge
    def mimic_trajectory(
        env,
        agent_uid: str,
        raw_edge: Dict,
        raw_affordance: Dict,
        target_edge: Dict,
        vis: bool = False,
        **kwargs,
    ):

        GeneralActionBank.generate_trajectory_qpos = generate_trajectory_qpos
        if isinstance(raw_affordance, dict):
            aff_kwargs = deepcopy(raw_affordance.get("kwargs", {}))
            aff_kwargs.pop("trajectory", {})
            aff_kwargs.pop("canonical_trajectory", {})
            getattr(GeneralActionBank, raw_affordance["name"])(
                env,
                **aff_kwargs,
                trajectory=env.trajectory,
                canonical_trajectory=env.canonical_trajectory,
            )
            xpos = env.affordance_datas[
                get_xpos_name(to_affordance_name(raw_affordance["name"]))
            ]
        else:
            log_warning(
                f"raw_affordance is not a dict, but {type(raw_affordance)} and {raw_affordance}. Using it as a string name directly."
            )
            xpos = env.affordance_datas[get_xpos_name(raw_affordance)]

        # raw_trajectory = getattr(GeneralActionBank, raw_edge["name"])(
        #     env, **raw_edge.get("kwargs", {})
        # )
        # base_pose = env.agent.get_base_xpos(agent_uid)

        # import time
        # for t, temp in enumerate([env.agent.get_fk(raw_trajectory[:, i], uid=agent_uid)
        #     for i in range(raw_trajectory.shape[1])
        # ]):
        #     if t % 10 ==0:

        #         print(temp)
        # env.scene.draw_marker(cfg=MarkerCfg(
        #     marker_type="axis",
        #     axis_xpos=env.agent.get_base_xpos(agent_uid) @ temp,
        #     axis_size=0.002,
        #     axis_len=0.005
        # ))
        #         time.sleep(0.01)

        # trans = np.linalg.inv(base_pose) @ new_xpos @ np.linalg.inv(xpos) @ base_pose
        # ref_poses = [
        #     trans @ env.agent.get_fk(raw_trajectory[:, i], uid=agent_uid)
        #     for i in range(raw_trajectory.shape[1])
        # ]
        # if vis:
        # env.scene.draw_marker(cfg=MarkerCfg(
        #     marker_type="axis",
        #     axis_xpos=xpos,
        #     axis_size=0.002,
        #     axis_len=0.005
        # ))

        # for t, temp in enumerate(ref_poses):
        #     print(temp)
        #     if t % 10 ==0:
        # env.scene.draw_marker(cfg=MarkerCfg(
        #     marker_type="axis",
        #     axis_xpos=env.agent.get_base_xpos(agent_uid) @ temp,
        #     axis_size=0.002,
        #     axis_len=0.005
        # ))
        #         time.sleep(0.01)

        ref_poses = []
        target_edge["name"] = "plan_trajectory"
        return getattr(GeneralActionBank, target_edge["name"])(
            env,
            ref_poses=ref_poses,
            duration=target_edge["duration"],
            vis=vis,
            **target_edge.get("kwargs", {}),
        )

    @staticmethod
    @tag_edge
    def plan_trajectory(
        env,
        agent_uid: str,
        keypose_names: List[str],
        duration: int,
        ref_poses: List[np.ndarray] = [],
        vis: bool = False,
        **kwargs,
    ) -> np.ndarray:
        from embodichain.lab.sim.planners import (
            MoveType,
            PlanState,
            MotionGenerator,
            MotionGenCfg,
            MotionGenOptions,
            ToppraPlanOptions,
            ToppraPlannerCfg,
        )

        # Retrieve the start and end positions
        start_qpos = env.affordance_datas[keypose_names[0]]

        control_part = get_control_part(env, agent_uid)
        start_qpos = torch.as_tensor(env.affordance_datas[keypose_names[0]])[None]
        start_xpos = torch.bmm(
            env.robot.get_control_part_base_pose(control_part, to_matrix=True),
            env.robot.compute_fk(start_qpos, control_part, to_matrix=True),
        )

        end_qpos = torch.as_tensor(env.affordance_datas[keypose_names[-1]])
        end_xpos = torch.bmm(
            env.robot.get_control_part_base_pose(control_part, to_matrix=True),
            env.robot.compute_fk(end_qpos, control_part, to_matrix=True),
        )

        # TODO: only 1 env
        start_qpos = start_qpos.squeeze(0).cpu().numpy()
        start_xpos = start_xpos.squeeze(0).cpu().numpy()
        end_qpos = end_qpos.squeeze(0).cpu().numpy()
        end_xpos = end_xpos.squeeze(0).cpu().numpy()

        if vis:
            env.sim.draw_marker(
                cfg=MarkerCfg(
                    marker_type="axis",
                    axis_xpos=start_xpos,
                    axis_size=0.002,
                    axis_len=0.005,
                )
            )

            env.sim.draw_marker(
                cfg=MarkerCfg(
                    marker_type="axis",
                    axis_xpos=end_xpos,
                    axis_size=0.002,
                    axis_len=0.005,
                )
            )

        filtered_keyposes = [start_qpos, end_qpos]
        if "eef" in agent_uid:
            filtered_keyposes = [start_qpos]

        if len(filtered_keyposes) == 1 and len(ref_poses) == 0:

            ret = np.array([filtered_keyposes[0]] * duration)
        else:
            mo_gen = MotionGenerator(
                cfg=MotionGenCfg(planner_cfg=ToppraPlannerCfg(robot_uid=env.robot.uid))
            )

            if len(ref_poses) == 0:
                plan_state = [
                    PlanState(qpos=torch.as_tensor(qpos), move_type=MoveType.JOINT_MOVE)
                    for qpos in filtered_keyposes
                ]

                ret = mo_gen.generate(
                    target_states=plan_state,
                    options=MotionGenOptions(
                        control_part=agent_uid,
                        plan_opts=ToppraPlanOptions(
                            sample_interval=duration,
                        ),
                    ),
                )

            else:
                plan_state = [
                    PlanState(
                        xpos=torch.as_tensor(start_xpos), move_type=MoveType.EEF_MOVE
                    )
                ]
                plan_state += [
                    PlanState(
                        xpos=torch.as_tensor(ref_pose), move_type=MoveType.EEF_MOVE
                    )
                    for ref_pose in ref_poses
                ]
                plan_state += [
                    PlanState(
                        xpos=torch.as_tensor(end_xpos), move_type=MoveType.EEF_MOVE
                    )
                ]

                ret = mo_gen.generate(
                    target_states=plan_state,
                    options=MotionGenOptions(
                        control_part=agent_uid,
                        is_interpolate=True,
                        plan_opts=ToppraPlanOptions(
                            sample_interval=duration,
                        ),
                    ),
                )

        return ret.positions.numpy().T

    @staticmethod
    @tag_edge
    def execute_open(env, **kwargs):
        from embodichain.lab.gym.utils.misc import (
            mul_linear_expand,
        )

        duration = kwargs.get("duration", 1)
        expand = kwargs.get("expand", True)
        if expand:
            action = mul_linear_expand(np.array([[1.0], [0.0]]), [duration - 1])
            action = np.concatenate([action, np.array([[0.0]])]).transpose()
        else:
            action = np.zeros((1, duration))
        return action

    @staticmethod
    @tag_edge
    def execute_close(env, **kwargs):
        from embodichain.lab.gym.utils.misc import (
            mul_linear_expand,
        )

        duration = kwargs.get("duration", 1)
        expand = kwargs.get("expand", True)
        if expand:
            action = mul_linear_expand(np.array([[0.0], [1.0]]), [duration - 1])
            action = np.concatenate([action, np.array([[1.0]])]).transpose()
        else:
            action = np.ones((1, duration))
        return action


class ActionBankMimic:
    def __init__(self, action_banks: List[ActionBank], prob: float = 0.5) -> None:
        self.action_banks = action_banks
        self.keyword = "mimicable"
        self.prob = prob

    def mimic(self, id=None) -> ActionBank:
        if len(self.action_banks) == 1:
            return self.action_banks[0]

        if id is None:
            id = np.random.randint(len(self.action_banks))
        assert id < len(
            self.action_banks
        ), f"Invalid id {id}, should be less than {len(self.action_banks)}"

        acb = self.action_banks[id]
        ret_acb = deepcopy(acb)
        node_names = acb.get_node_names(bool_attr_name=self.keyword)

        ret_node_grap2id = ret_acb.graph2id()
        edge_need_modify = {}
        for scope in node_names.keys():
            edge_need_modify[scope] = []
            # if np.random.random() < self.prob:
            #     continue
            for node in node_names[scope]:
                mimic_id = np.random.randint(len(self.action_banks))
                action_bank_mimic_for_this_node = self.action_banks[mimic_id]
                mimic_node_names = action_bank_mimic_for_this_node.get_node_names(
                    bool_attr_name=self.keyword
                )
                temp_graph2id = action_bank_mimic_for_this_node.graph2id()

                if node in mimic_node_names[scope]:
                    ret_acb.conf["node"][scope][ret_node_grap2id[scope][node]][node] = (
                        deepcopy(
                            action_bank_mimic_for_this_node.conf["node"][scope][
                                temp_graph2id[scope][node]
                            ][node]
                        )
                    )

                    edges = ret_acb.get_edge_names(node_name=node)
                    for edge in edges[scope]:
                        edge.update({"mimic_id": mimic_id})
                    edge_need_modify[scope].extend(edges[scope])
                else:
                    log_warning(
                        f"Node {node} in scope {scope} not found in action bank {mimic_id} [{mimic_node_names[scope]}] to mimic."
                    )
            edge_need_modify[scope] = {
                v["name"]: v for v in edge_need_modify[scope]
            }.values()

        raw_edge_grap2id = acb.graph2id("edge")
        raw_node_grap2id = acb.graph2id("node")
        ret_edge_grap2id = ret_acb.graph2id("edge")
        for scope in node_names.keys():
            for edge in edge_need_modify[scope]:
                edge_name = edge["name"]
                mimic_id = edge.pop("mimic_id")
                mimic_grap2id = self.action_banks[mimic_id].graph2id("edge")
                ret_acb.conf["edge"][scope][ret_edge_grap2id[scope][edge_name]][
                    edge_name
                ]["name"] = "mimic_trajectory"
                ret_acb.conf["edge"][scope][ret_edge_grap2id[scope][edge_name]][
                    edge_name
                ]["duration"] = deepcopy(
                    self.action_banks[mimic_id].conf["edge"][scope][
                        mimic_grap2id[scope][edge_name]
                    ][edge_name]["duration"]
                )

                ret_acb.conf["edge"][scope][ret_edge_grap2id[scope][edge_name]][
                    edge_name
                ]["kwargs"] = {}
                ret_acb.conf["edge"][scope][ret_edge_grap2id[scope][edge_name]][
                    edge_name
                ]["kwargs"]["agent_uid"] = scope
                ret_acb.conf["edge"][scope][ret_edge_grap2id[scope][edge_name]][
                    edge_name
                ]["kwargs"]["raw_edge"] = deepcopy(
                    acb.conf["edge"][scope][raw_edge_grap2id[scope][edge_name]][
                        edge_name
                    ]
                )
                ret_acb.conf["edge"][scope][ret_edge_grap2id[scope][edge_name]][
                    edge_name
                ]["kwargs"]["raw_affordance"] = (
                    deepcopy(
                        acb.conf["node"][scope][raw_node_grap2id[scope][node]][node]
                    )
                    if node in raw_node_grap2id[scope]
                    else node
                )
                ret_acb.conf["edge"][scope][ret_edge_grap2id[scope][edge_name]][
                    edge_name
                ]["kwargs"]["target_edge"] = deepcopy(
                    self.action_banks[mimic_id].conf["edge"][scope][
                        mimic_grap2id[scope][edge_name]
                    ][edge_name]
                )

        return ret_acb



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import warp as wp
from typing import Any


@wp.kernel(enable_backward=False)
def reshape_tiled_image(
    tiled_image_buffer: Any,
    batched_image: Any,
    image_height: int,
    image_width: int,
    num_channels: int,
    num_tiles_x: int,
):
    """Reshapes a tiled image into a batch of images.

    This function reshapes the input tiled image buffer into a batch of images. The input image buffer
    is assumed to be tiled in the x and y directions. The output image is a batch of images with the
    specified height, width, and number of channels.

    Args:
        tiled_image_buffer: The input image buffer. Shape is (height * width * num_channels * num_cameras,).
        batched_image: The output image. Shape is (num_cameras, height, width, num_channels).
        image_width: The width of the image.
        image_height: The height of the image.
        num_channels: The number of channels in the image.
        num_tiles_x: The number of tiles in x-direction.
    """
    # get the thread id
    camera_id, height_id, width_id = wp.tid()

    # resolve the tile indices
    tile_x_id = camera_id % num_tiles_x
    # TODO: Currently, the tiles arranged in the bottom-to-top order, which should be changed.
    tile_y_id = (
        num_tiles_x - 1 - (camera_id // num_tiles_x)
    )  # Adjust for bottom-to-top tiling
    # compute the start index of the pixel in the tiled image buffer
    pixel_start = (
        num_channels
        * num_tiles_x
        * image_width
        * (image_height * tile_y_id + height_id)
        + num_channels * tile_x_id * image_width
        + num_channels * width_id
    )

    # copy the pixel values into the batched image
    for i in range(num_channels):
        batched_image[camera_id, height_id, width_id, i] = batched_image.dtype(
            tiled_image_buffer[pixel_start + i]
        )


# uint32 -> int32 conversion is required for non-colored segmentation annotators
wp.overload(
    reshape_tiled_image,
    {
        "tiled_image_buffer": wp.array(dtype=wp.uint32),
        "batched_image": wp.array(dtype=wp.uint32, ndim=4),
    },
)
# uint8 is used for 4 channel annotators
wp.overload(
    reshape_tiled_image,
    {
        "tiled_image_buffer": wp.array(dtype=wp.uint8),
        "batched_image": wp.array(dtype=wp.uint8, ndim=4),
    },
)
# float32 is used for single channel annotators
wp.overload(
    reshape_tiled_image,
    {
        "tiled_image_buffer": wp.array(dtype=wp.float32),
        "batched_image": wp.array(dtype=wp.float32, ndim=4),
    },
)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from __future__ import annotations

import dexsim
import math
import torch
import dexsim.render as dr
import warp as wp

from functools import cached_property
from typing import Union, Tuple, Sequence, List

from embodichain.lab.sim.sensors import BaseSensor, SensorCfg
from embodichain.utils.math import matrix_from_quat, quat_from_matrix, look_at_to_pose
from embodichain.utils.warp.kernels import reshape_tiled_image
from embodichain.utils import logger, configclass
from embodichain.lab.sim.utility.sim_utils import is_rt_enabled


@configclass
class CameraCfg(SensorCfg):
    """Configuration class for Camera."""

    @configclass
    class ExtrinsicsCfg(SensorCfg.OffsetCfg):
        """Configuration class for camera extrinsics.

        The extrinsics define the position and orientation of the camera in the 3D world.
        If eye, target, and up are provided, they will be used to compute the extrinsics.
        Otherwise, the position and orientation will be set to the defaults.
        """

        eye: Tuple[float, float, float] | None = None
        target: Tuple[float, float, float] | None = None
        up: Tuple[float, float, float] | None = None
        """Alternative way to specify the camera extrinsics using eye, target, and up vectors."""

        @property
        def transformation(self) -> torch.Tensor:
            if self.eye:
                self.up = (0.0, 0.0, 1.0) if self.up is None else self.up
                return look_at_to_pose(self.eye, self.target, self.up).squeeze(0)
            else:
                return super().transformation

    sensor_type: str = "Camera"

    # Camera parameters
    width: int = 640
    height: int = 480
    near: float = 0.005
    far: float = 100.0

    # The camera intrinsics are defined as (fx, fy, cx, cy)
    intrinsics: Tuple[float, float, float, float] = (600, 600, 320.0, 240.0)
    extrinsics: ExtrinsicsCfg = ExtrinsicsCfg()

    enable_color: bool = True
    enable_depth: bool = False
    enable_mask: bool = False
    enable_normal: bool = False
    enable_position: bool = False

    fx: float = intrinsics[0]
    fy: float = intrinsics[1]
    cx: float = intrinsics[2]
    cy: float = intrinsics[3]

    def get_view_attrib(self) -> dr.ViewFlags:
        """Get the view attributes for the camera.

        The camera view whcich is used to render the scene
        Default view attributes for the camera are: [COLOR, DEPTH, MASK]
        The supported view attributes are:
            - COLOR: RGBA images
            - DEPTH: Depth images
            - MASK: Instance segmentation masks
            - NORMAL: Normal images
            - POSITION: Position images with 3D coordinates.

        Returns:
            The view attributes for the camera.
        """
        view_attrib: dr.ViewFlags = dr.ViewFlags.COLOR
        # TODO: change for fast-rt renderer backend.
        if self.enable_color:
            view_attrib |= dr.ViewFlags.COLOR
        if self.enable_depth:
            if is_rt_enabled() is False:
                view_attrib |= dr.ViewFlags.NORMAL
            view_attrib |= dr.ViewFlags.DEPTH
        if self.enable_mask:
            view_attrib |= dr.ViewFlags.MASK
            if is_rt_enabled() is False:
                view_attrib |= dr.ViewFlags.DEPTH
        if self.enable_normal:
            view_attrib |= dr.ViewFlags.NORMAL
        if self.enable_position:
            view_attrib |= dr.ViewFlags.POSITION
        return view_attrib

    def get_data_types(self) -> List[str]:
        data_types = []
        if self.enable_color:
            data_types.append("color")
        if self.enable_depth:
            data_types.append("depth")
        if self.enable_mask:
            data_types.append("mask")
        if self.enable_normal:
            data_types.append("normal")
        if self.enable_position:
            data_types.append("position")
        return data_types


class Camera(BaseSensor):
    """Base class for sensor abstraction in the simulation engine.

    Sensors should inherit from this class and implement the `update` and `get_data` methods.
    """

    SUPPORTED_DATA_TYPES = ["color", "depth", "mask", "normal", "position"]

    def __init__(
        self, config: CameraCfg, device: torch.device = torch.device("cpu")
    ) -> None:
        super().__init__(config, device)

    def _build_sensor_from_config(
        self, config: CameraCfg, device: torch.device
    ) -> None:
        self._world = dexsim.default_world()
        env = self._world.get_env()
        arenas = env.get_all_arenas()
        if len(arenas) == 0:
            arenas = [env]
        num_instances = len(arenas)

        if self.is_rt_enabled:
            self._frame_buffer = self._world.create_camera_group(
                [config.width, config.height], num_instances, True
            )

            view_attrib = config.get_view_attrib()
            for i, arena in enumerate(arenas):
                view_name = f"{self.uid}_view{i + 1}"
                view = arena.create_camera(
                    view_name,
                    config.width,
                    config.height,
                    True,
                    view_attrib,
                    self._frame_buffer,
                )
                view.set_intrinsic(config.intrinsics)
                view.set_near(config.near)
                view.set_far(config.far)
                self._entities[i] = view

        else:
            self._grid_size = math.ceil(math.sqrt(num_instances))
            frame_width = self._grid_size * config.width
            frame_height = self._grid_size * config.height
            view_attrib = config.get_view_attrib()
            # Create the data frame
            self._frame_buffer = self._world.create_frame_buffer(
                [frame_width, frame_height], view_attrib, True
            )
            self._frame_buffer.set_read_able(view_attrib)

            # Create camera views
            for i, arena in enumerate(arenas):
                col = i // self._grid_size
                row = i % self._grid_size
                x = row * config.width
                y = col * config.height
                view_name = f"{self.uid}_view{i + 1}"

                view = arena.create_camera_view(
                    view_name, (x, y), (config.width, config.height), self._frame_buffer
                )
                view.set_intrinsic(config.intrinsics)
                view.set_near(config.near)
                view.set_far(config.far)
                view.enable_postprocessing(True)

                self._entities[i] = view

        # Define a mapping of data types to their respective shapes and dtypes
        buffer_specs = {
            "color": (
                (self.num_instances, config.height, config.width, 4),
                torch.uint8,
            ),
            "depth": (
                (self.num_instances, config.height, config.width),
                torch.float32,
            ),
            "mask": (
                (self.num_instances, config.height, config.width),
                torch.int32,
            ),
            "normal": (
                (self.num_instances, config.height, config.width, 3),
                torch.float32,
            ),
            "position": (
                (self.num_instances, config.height, config.width, 3),
                torch.float32,
            ),
        }
        data_types = config.get_data_types()

        # Iterate through enabled data types and initialize buffers
        for data_type in data_types:
            if getattr(config, f"enable_{data_type}", False):
                shape, dtype = buffer_specs[data_type]
                self._data_buffer[data_type] = torch.empty(
                    shape, dtype=dtype, device=device
                )

        self.cfg: CameraCfg = config
        if self.cfg.extrinsics.parent is not None:
            self._attach_to_entity()

    @cached_property
    def is_rt_enabled(self) -> bool:
        """Check if Ray Tracing rendering backend is enabled in the default dexsim world.

        Returns:
            bool: True if Ray Tracing rendering is enabled, False otherwise.
        """
        return is_rt_enabled()

    @cached_property
    def group_id(self) -> int:
        """Get the camera group ID in the dexsim world.

        Returns:
            int: The camera group ID.
        """
        if self.is_rt_enabled:
            return self._frame_buffer.get_group_id()
        else:
            logger.log_warning(
                "Camera group ID is only available for Ray Tracing renderer. Returning -1 for non-RT renderer."
            )
            return -1

    @property
    def is_attached(self) -> bool:
        """Check if the camera is attached to a parent entity.

        Returns:
            bool: True if the camera is attached to a parent entity, False otherwise.
        """
        return self.cfg.extrinsics.parent is not None

    def update(self, **kwargs) -> None:
        """Update the sensor data.

        The supported data types are:
            - color: RGB images with shape (B, H, W, 4) and dtype torch.uint8
            - depth: Depth images with shape (B, H, W) and dtype torch.float32
            - mask: Instance segmentation masks with shape (B, H, W) and dtype torch.int32
            - normal: Normal images with shape (B, H, W, 3) and dtype torch.float32
            - position: Position images with shape (B, H, W, 3) and dtype torch.float32

        Args:
            **kwargs: Additional keyword arguments for sensor update.
                - fetch_only (bool): If True, only fetch the data from dexsim internal frame buffer without performing rendering.
        """
        fetch_only = kwargs.get("fetch_only", False)
        if not fetch_only:
            if self.is_rt_enabled:
                self._frame_buffer.apply()
            else:
                self._frame_buffer.apply_frame()

        self.cfg: CameraCfg
        # TODO: support fetch data from gpu buffer directly.
        if self.cfg.enable_color:
            if self.is_rt_enabled:
                self._data_buffer["color"] = self._frame_buffer.get_rgb_gpu_buffer().to(
                    self.device
                )
            else:
                data = self._frame_buffer.get_color_gpu_buffer().to(self.device)
                self._update_buffer_impl(data, self._data_buffer["color"])

        if self.cfg.enable_depth:
            data = self._frame_buffer.get_depth_gpu_buffer().to(self.device)
            if self.is_rt_enabled:
                self._data_buffer["depth"] = data
            else:
                self._update_buffer_impl(
                    data, self._data_buffer["depth"].unsqueeze_(-1)
                )
                self._data_buffer["depth"].squeeze_(-1)

        if self.cfg.enable_mask:
            if self.is_rt_enabled:
                data = self._frame_buffer.get_visible_mask_gpu_buffer().to(
                    self.device, torch.int32
                )
                self._data_buffer["mask"] = data
            else:
                data = self._frame_buffer.get_visible_gpu_buffer().to(
                    self.device, torch.int32
                )
                self._update_buffer_impl(data, self._data_buffer["mask"].unsqueeze_(-1))
                self._data_buffer["mask"].squeeze_(-1)

        if self.cfg.enable_normal:
            data = self._frame_buffer.get_normal_gpu_buffer().to(self.device)
            if self.is_rt_enabled:
                self._data_buffer["normal"] = data
            else:
                self._update_buffer_impl(data, self._data_buffer["normal"])

        if self.cfg.enable_position:
            data = self._frame_buffer.get_position_gpu_buffer().to(self.device)
            if self.is_rt_enabled:
                self._data_buffer["position"] = data
            else:
                self._update_buffer_impl(data, self._data_buffer["position"])

    def _update_buffer_impl(
        self, data_buffer: torch.Tensor, data_buffer_out: torch.Tensor
    ) -> None:
        device = str(self.device)
        channel = data_buffer.shape[-1] if data_buffer.dim() >= 3 else 1
        wp.launch(
            kernel=reshape_tiled_image,
            dim=(self.num_instances, self.cfg.height, self.cfg.width),
            inputs=[
                wp.from_torch(data_buffer).flatten(),
                wp.from_torch(data_buffer_out),
                self.cfg.height,
                self.cfg.width,
                channel,
                self._grid_size,
            ],
            device="cuda:0" if device == "cuda" else device,
        )

    def _attach_to_entity(self) -> None:
        """Attach the sensor to the parent entity in each environment."""
        env = self._world.get_env()
        for i, entity in enumerate(self._entities):

            parent = None
            if i == 0:
                parent = env.find_node(f"{self.cfg.extrinsics.parent}")
            else:
                parent = env.find_node(f"{self.cfg.extrinsics.parent}.{i-1}")
            if parent is None:
                logger.log_error(
                    f"Failed to find parent entity {self.cfg.extrinsics.parent} for sensor {self.cfg.uid}."
                )

            entity.attach_node(parent)

    def set_local_pose(
        self, pose: torch.Tensor, env_ids: Sequence[int] | None = None
    ) -> None:
        """Set the local pose of the camera.

        Note: The pose should be in the OpenGL coordinate system, which means the Y is up and Z is forward.

        Args:
            pose (torch.Tensor): The local pose to set, should be a 4x4 transformation matrix.
            env_ids (Sequence[int] | None): The environment IDs to set the pose for. If None, set for all environments.
        """
        if env_ids is None:
            local_env_ids = range(len(self._entities))
        else:
            local_env_ids = env_ids

        pose = pose.cpu()
        if pose.dim() == 2 and pose.shape[1] == 7:
            pose_matrix = torch.eye(4).unsqueeze(0).repeat(pose.shape[0], 1, 1)
            pose_matrix[:, :3, 3] = pose[:, :3]
            pose_matrix[:, :3, :3] = matrix_from_quat(pose[:, 3:7])
            for i, env_idx in enumerate(local_env_ids):
                self._entities[env_idx].set_local_pose(pose_matrix[i].numpy())
        elif pose.dim() == 3 and pose.shape[1:] == (4, 4):
            for i, env_idx in enumerate(local_env_ids):
                self._entities[env_idx].set_local_pose(pose[i].numpy())
        else:
            logger.log_error(
                f"Invalid pose shape {pose.shape}. Expected (N, 7) or (N, 4, 4)."
            )

    def get_local_pose(self, to_matrix: bool = False) -> torch.Tensor:
        """Get the local pose of the camera.

        Args:
            to_matrix (bool): If True, return the pose as a 4x4 matrix. If False, return as a quaternion.

        Returns:
            torch.Tensor: The local pose of the camera.
        """
        poses = []
        for entity in self._entities:
            pose = entity.get_local_pose()
            poses.append(torch.as_tensor(pose, dtype=torch.float32))

        poses = torch.stack(poses, dim=0).to(self.device)
        if to_matrix is False:
            xyz = poses[:, :3, 3]
            quat = quat_from_matrix(poses[:, :3, :3])
            return torch.cat((xyz, quat), dim=-1)
        return poses

    def get_arena_pose(self, to_matrix: bool = False) -> torch.Tensor:
        """Get the pose of the sensor in the arena frame.

        Args:
            to_matrix (bool): If True, return the pose as a 4x4 transformation matrix.

        Returns:
            A tensor representing the pose of the sensor in the arena frame.
        """
        from embodichain.lab.sim.utility import get_dexsim_arenas

        arenas = get_dexsim_arenas()

        poses = []
        for i, entity in enumerate(self._entities):
            pose = entity.get_world_pose()
            pose[:2, 3] -= arenas[i].get_root_node().get_local_pose()[:2, 3]
            poses.append(torch.as_tensor(pose, dtype=torch.float32))

        poses = torch.stack(poses, dim=0).to(self.device)
        if to_matrix is False:
            xyz = poses[:, :3, 3]
            quat = quat_from_matrix(poses[:, :3, :3])
            return torch.cat((xyz, quat), dim=-1)
        return poses

    def look_at(
        self,
        eye: torch.Tensor,
        target: torch.Tensor,
        up: torch.Tensor | None = None,
        env_ids: Sequence[int] | None = None,
    ) -> None:
        """Set the camera to look at a target point.

        Args:
            eye (torch.Tensor): The position of the camera (eye) with shape (N, 3).
            target (torch.Tensor): The point the camera should look at (target) with shape (N, 3).
            up (torch.Tensor | None): The up direction vector. If None, defaults to [0, 0, 1].
            env_ids (Sequence[int] | None): The environment IDs to set the look at for. If None, set for all environments.
        """
        if up is None:
            up = torch.tensor([[0.0, 0.0, 1.0]]).repeat(eye.shape[0], 1)

        pose = look_at_to_pose(eye, target, up)
        # To opengl coordinate system.
        pose[:, :3, 1] = -pose[:, :3, 1]
        pose[:, :3, 2] = -pose[:, :3, 2]
        self.set_local_pose(pose, env_ids=env_ids)

    def set_intrinsics(
        self,
        intrinsics: torch.Tensor,
        env_ids: Sequence[int] | None = None,
    ) -> None:
        """
        Set the camera intrinsics for both left and right cameras.

        Args:
            intrinsics (torch.Tensor): The intrinsics for the left camera with shape (4,) / (3, 3) or (N, 4) / (N, 3, 3).
            env_ids (Sequence[int] | None): The environment ids to set the intrinsics. If None, set for all environments.
        """
        ids = env_ids if env_ids is not None else range(self.num_instances)

        if intrinsics.dim() == 2 and intrinsics.shape[1] == 3:
            intrinsics = intrinsics.unsqueeze(0).repeat(len(ids), 1, 1)

        if intrinsics.dim() == 1:
            intrinsics = intrinsics.unsqueeze(0).repeat(len(ids), 1)

        if len(ids) != intrinsics.shape[0]:
            logger.log_error(
                f"Invalid intrinsics shape {intrinsics.shape} for {len(ids)} environments."
            )

        for i, env_id in enumerate(ids):
            entity = self._entities[env_id]
            if intrinsics.shape[1] == 3:
                entity.set_intrinsic(intrinsics[i].cpu().numpy())
            else:
                entity.set_intrinsic(intrinsics[i].cpu().tolist())

    def get_intrinsics(self) -> torch.Tensor:
        """
        Get the camera intrinsics for both left and right cameras.

        Returns:
            torch.Tensor: The intrinsics for the left camera with shape (N, 3, 3).
        """
        intrinsics = []
        for entity in self._entities:
            intrinsics.append(
                torch.as_tensor(entity.get_intrinsic(), dtype=torch.float32)
            )

        return torch.stack(intrinsics, dim=0).to(self.device)

    def reset(self, env_ids: Sequence[int] | None = None) -> None:
        self.cfg: CameraCfg

        if self.cfg.extrinsics.eye is not None:
            eye = (
                torch.tensor(self.cfg.extrinsics.eye, dtype=torch.float32)
                .squeeze_(0)
                .repeat(self.num_instances, 1)
            )
            target = (
                torch.tensor(self.cfg.extrinsics.target, dtype=torch.float32)
                .squeeze_(0)
                .repeat(self.num_instances, 1)
            )
            up = (
                torch.tensor(self.cfg.extrinsics.up, dtype=torch.float32)
                .squeeze_(0)
                .repeat(self.num_instances, 1)
                if self.cfg.extrinsics.up is not None
                else None
            )
            self.look_at(eye, target, up, env_ids=env_ids)
        else:
            pose = self.cfg.extrinsics.transformation
            pose = pose.unsqueeze_(0).repeat(self.num_instances, 1, 1)

            if self.cfg.extrinsics.parent is None:
                # To opengl coordinate system.
                pose[:, :3, 1] = -pose[:, :3, 1]
                pose[:, :3, 2] = -pose[:, :3, 2]

            self.set_local_pose(pose, env_ids=env_ids)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.

# All rights reserved.
#
# This file incorporates code from the Isaac Lab Project
# Copyright (c) 2022-2025, The Isaac Lab Project Developers
# (https://github.com/isaac-sim/IsaacLab/blob/main/CONTRIBUTORS.md).
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause
# ----------------------------------------------------------------------------

"""Event manager for orchestrating operations based on different simulation events."""

from __future__ import annotations

import inspect
import torch
from collections.abc import Sequence
from prettytable import PrettyTable
from typing import TYPE_CHECKING, Union

from embodichain.utils import logger
from .manager_base import ManagerBase
from .cfg import EventCfg

if TYPE_CHECKING:
    from embodichain.lab.gym.envs import EmbodiedEnv


class EventManager(ManagerBase):
    """Manager for orchestrating operations based on different simulation events.

    The event manager applies operations to the environment based on different simulation events. For example,
    changing the masses of objects or their friction coefficients during initialization/ reset, or applying random
    pushes to the robot at a fixed interval of steps. The user can specify several modes of events to fine-tune the
    behavior based on when to apply the event.

    The event functors are parsed from a config class containing the manager's settings and each functor's
    parameters. Each event functor should instantiate the :class:`EventCfg` class.

    Event functors can be grouped by their mode. The mode is a user-defined string that specifies when
    the event functor should be applied. This provides the user complete control over when event
    functors should be applied.

    For a typical training process, you may want to apply events in the following modes:

    - "prestartup": Event is applied once at the beginning of the training before the simulation starts.
      This is used to randomize USD-level properties of the simulation stage.
    - "startup": Event is applied once at the beginning of the training once simulation is started.
    - "reset": Event is applied at every reset.
    - "interval": Event is applied at pre-specified intervals of time.

    However, you can also define your own modes and use them in the training process as you see fit.
    For this you will need to add the triggering of that mode in the environment implementation as well.

    .. note::

        The triggering of operations corresponding to the mode ``"interval"`` are the only mode that are
        directly handled by the manager itself. The other modes are handled by the environment implementation.

    """

    _env: EmbodiedEnv
    """The environment instance."""

    def __init__(self, cfg: object, env: EmbodiedEnv):
        """Initialize the event manager.

        Args:
            cfg: A configuration object or dictionary (``dict[str, EventCfg]``).
            env: An environment object.
        """
        # create buffers to parse and store functors
        self._mode_functor_names: dict[str, list[str]] = dict()
        self._mode_functor_cfgs: dict[str, list[EventCfg]] = dict()
        self._mode_class_functor_cfgs: dict[str, list[EventCfg]] = dict()

        # call the base class (this will parse the functors config)
        super().__init__(cfg, env)

    def __str__(self) -> str:
        """Returns: A string representation for event manager."""
        functor_num = sum(len(v) for v in self._mode_functor_names.values())
        msg = f"<EventManager> contains {functor_num} active functors.\n"

        # add info on each mode
        for mode in self._mode_functor_names:
            # create table for functor information
            table = PrettyTable()
            table.title = f"Active Event Functors in Mode: '{mode}'"
            # add table headers based on mode
            if mode == "interval":
                table.field_names = ["Index", "Name", "Interval step"]
                table.align["Name"] = "l"
                for index, (name, cfg) in enumerate(
                    zip(self._mode_functor_names[mode], self._mode_functor_cfgs[mode])
                ):
                    table.add_row([index, name, cfg.interval_step])
            else:
                table.field_names = ["Index", "Name"]
                table.align["Name"] = "l"
                for index, name in enumerate(self._mode_functor_names[mode]):
                    table.add_row([index, name])
            # convert table to string
            msg += table.get_string()
            msg += "\n"

        return msg

    """
    Properties.
    """

    @property
    def active_functors(self) -> dict[str, list[str]]:
        """Name of active event functors.

        The keys are the modes of event and the values are the names of the event functors.
        """
        return self._mode_functor_names

    @property
    def available_modes(self) -> list[str]:
        """Modes of events."""
        return list(self._mode_functor_names.keys())

    """
    Operations.
    """

    def reset(self, env_ids: Union[Sequence[int], None] = None) -> dict[str, float]:
        # call all functors that are classes
        for mode_cfg in self._mode_class_functor_cfgs.values():
            for functor_cfg in mode_cfg:
                functor_cfg.func.reset(env_ids=env_ids)

        # resolve number of environments
        if env_ids is None:
            num_envs = self._env.num_envs
        else:
            num_envs = len(env_ids)

        # May be add more useful reset logic later.

        # nothing to log here
        return {}

    def apply(
        self,
        mode: str,
        env_ids: Union[Sequence[int], None] = None,
    ):
        """Calls each event functor in the specified mode.

        This function iterates over all the event functors in the specified mode and calls the function
        corresponding to the functor. The function is called with the environment instance and the environment
        indices to apply the event to.

        For the "interval" mode, the function is called when the time interval has passed. This requires
        specifying the time step of the environment.

        For the "reset" mode, the function is called when the mode is "reset" and the total number of environment
        steps that have happened since the last trigger of the function is equal to its configured parameter for
        the number of environment steps between resets.

        Args:
            mode: The mode of event.
            env_ids: The indices of the environments to apply the event to.
                Defaults to None, in which case the event is applied to all environments when applicable.

        Raises:
            ValueError: If the mode is ``"interval"`` and the environment indices are provided. This is an undefined
                behavior as the environment indices are computed based on the time left for each environment.
            ValueError: If the mode is ``"reset"`` and the total number of environment steps that have happened
                is not provided.
        """
        # check if mode is valid
        if mode not in self._mode_functor_names:
            logger.log_warning(f"Event mode '{mode}' is not defined. Skipping event.")
            return

        if mode == "interval" and env_ids is not None:
            logger.log_error(
                f"Event mode '{mode}' does not require environment indices. This is an undefined behavior"
                " as the environment indices are computed based on the time left for each environment."
            )

        # iterate over all the event functors
        for index, functor_cfg in enumerate(self._mode_functor_cfgs[mode]):
            functor_cfg: EventCfg
            if mode == "interval":
                self._interval_functor_step_count[index] += 1

                # check if the interval has passed and sample a new interval
                # note: we compare with a small value to handle floating point errors
                if (
                    functor_cfg.is_global
                    and self._interval_functor_step_count[index]
                    % functor_cfg.interval_step
                    == 0
                ):

                    # call the event functor (with None for env_ids)
                    functor_cfg.func(self._env, None, **functor_cfg.params)
                else:
                    valid_env_ids = (
                        (
                            self._interval_functor_step_count[index]
                            % functor_cfg.interval_step
                            == 0
                        )
                        .nonzero()
                        .flatten()
                    )
                    if len(valid_env_ids) > 0:
                        # call the event functor
                        functor_cfg.func(self._env, valid_env_ids, **functor_cfg.params)
            elif mode == "reset":
                # resolve the environment indices
                if env_ids is None:
                    env_ids = slice(None)

                functor_cfg.func(self._env, env_ids, **functor_cfg.params)
            else:
                # call the event functor
                functor_cfg.func(self._env, env_ids, **functor_cfg.params)

    """
    Operations - Functor settings.
    """

    def set_functor_cfg(self, functor_name: str, cfg: EventCfg):
        """Sets the configuration of the specified functor into the manager.

        The method finds the functor by name by searching through all the modes.
        It then updates the configuration of the functor with the first matching name.

        Args:
            functor_name: The name of the event functor.
            cfg: The configuration for the event functor.

        Raises:
            ValueError: If the functor name is not found.
        """
        functor_found = False
        for mode, functors in self._mode_functor_names.items():
            if functor_name in functors:
                self._mode_functor_cfgs[mode][functors.index(functor_name)] = cfg
                functor_found = True
                break
        if not functor_found:
            logger.log_error(f"Event functor '{functor_name}' not found.")

    def get_functor_cfg(self, functor_name: str) -> EventCfg:
        """Gets the configuration for the specified functor.

        The method finds the functor by name by searching through all the modes.
        It then returns the configuration of the functor with the first matching name.

        Args:
            functor_name: The name of the event functor.

        Returns:
            The configuration of the event functor.

        Raises:
            ValueError: If the functor name is not found.
        """
        for mode, functors in self._mode_functor_names.items():
            if functor_name in functors:
                return self._mode_functor_cfgs[mode][functors.index(functor_name)]
        logger.log_error(f"Event functor '{functor_name}' not found.")

    """
    Operations - Visit functor.
    """

    def get_functor(self, functor_name: str):
        """
        Retrieve a functor from the configuration by its name.

        Args:
            functor_name (str): The name of the functor to retrieve.

        Returns:
            The functor if it exists in the configuration, otherwise None.
        """
        if hasattr(self.cfg, functor_name):
            functor = getattr(self.cfg, functor_name).func
            return functor
        else:
            logger.log_warning(
                f"Got no functor {functor_name} in event_manager, please check again."
            )
            return None

    """
    Helper functions.
    """

    def _prepare_functors(self):
        # buffer to store the time left for "interval" mode
        # if interval is global, then it is a single value, otherwise it is per environment
        self._interval_functor_step_count: list[torch.Tensor] = list()

        # check if config is dict already
        if isinstance(self.cfg, dict):
            cfg_items = self.cfg.items()
        else:
            cfg_items = self.cfg.__dict__.items()
        # iterate over all the functors
        for functor_name, functor_cfg in cfg_items:
            # check for non config
            if functor_cfg is None:
                continue
            # check for valid config type
            if not isinstance(functor_cfg, EventCfg):
                raise TypeError(
                    f"Configuration for the functor '{functor_name}' is not of type EventCfg."
                    f" Received: '{type(functor_cfg)}'."
                )

            # resolve common parameters
            self._resolve_common_functor_cfg(functor_name, functor_cfg, min_argc=2)

            # check if mode is a new mode
            if functor_cfg.mode not in self._mode_functor_names:
                # add new mode
                self._mode_functor_names[functor_cfg.mode] = list()
                self._mode_functor_cfgs[functor_cfg.mode] = list()
                self._mode_class_functor_cfgs[functor_cfg.mode] = list()
            # add functor name and parameters
            self._mode_functor_names[functor_cfg.mode].append(functor_name)
            self._mode_functor_cfgs[functor_cfg.mode].append(functor_cfg)

            # check if the functor is a class
            if inspect.isclass(functor_cfg.func):
                self._mode_class_functor_cfgs[functor_cfg.mode].append(functor_cfg)

            # resolve the mode of the events
            # -- interval mode
            if functor_cfg.mode == "interval":
                # sample the time left for global
                if functor_cfg.is_global:
                    count = torch.zeros(1, dtype=torch.int32, device=self.device)
                    self._interval_functor_step_count.append(count)
                else:
                    count = torch.zeros(
                        self.num_envs, dtype=torch.int32, device=self.device
                    )
                    self._interval_functor_step_count.append(count)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import torch
import numpy as np
from IPython import embed

from embodichain.lab.sim import SimulationManager, SimulationManagerCfg
from embodichain.lab.sim.robots import DexforceW1Cfg

from embodichain.lab.sim.utility.workspace_analyzer.workspace_analyzer import (
    WorkspaceAnalyzer,
)

if __name__ == "__main__":
    # Example usage
    np.set_printoptions(precision=5, suppress=True)
    torch.set_printoptions(precision=5, sci_mode=False)

    config = SimulationManagerCfg(headless=False, sim_device="cpu")
    sim_manager = SimulationManager(config)
    sim_manager.set_manual_update(False)

    cfg = DexforceW1Cfg.from_dict(
        {"uid": "dexforce_w1", "version": "v021", "arm_kind": "industrial"}
    )
    robot = sim_manager.add_robot(cfg=cfg)
    print("DexforceW1 robot added to the simulation.")

    print("Example: Joint Space Analysis")

    wa_joint = WorkspaceAnalyzer(robot=robot, sim_manager=sim_manager)
    results_joint = wa_joint.analyze(num_samples=3000, visualize=True)

    print(f"\nJoint Space Results:")
    print(
        f"  Valid points: {results_joint['num_valid']} / {results_joint['num_samples']}"
    )
    print(f"  Analysis time: {results_joint['analysis_time']:.2f}s")
    print(f"  Metrics: {results_joint['metrics']}")

    embed(header="End of Joint Space Analysis Example")



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import abc  # for abstract base class definitions


class Device(metaclass=abc.ABCMeta):
    """
    Base class for all robot controllers.
    Defines basic interface for all controllers to adhere to.
    """

    @abc.abstractmethod
    def start_control(self):
        """
        Method that should be called externally before controller can
        start receiving commands.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def stop_control(self):
        """
        Method that should be called externally to stop the controller.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_controller_state(self):
        """Returns the current state of the device, a dictionary of pos, orn, grasp, and reset."""
        raise NotImplementedError



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import open3d as o3d
import numpy as np
import time
import pathlib
import pickle
import os

from enum import Enum
from copy import deepcopy
from typing import List

from .cone_sampler import ConeSampler
from embodichain.utils.utility import get_mesh_md5
from embodichain.utils import logger


class GraspSelectMethod(Enum):
    NORMAL_SCORE = 0
    NEAR_APPROACH = 1
    CENTER = 2


class AntipodalGrasp:
    def __init__(self, pose: np.ndarray, open_len: float, score: float) -> None:
        self.pose = pose  # [4, 4] of float grasp pose
        self.open_len = open_len  # gripper open length
        self.score = score  # grasp pose score

    def grasp_pose_visual_mesh(self, gripper_open_length: float = None):
        if gripper_open_length is None:
            gripper_open_length = self.open_len
            open_ratio = 1.0
        else:
            open_ratio = self.open_len / gripper_open_length
        open_ratio = max(1e-4, open_ratio)
        gripper_finger = o3d.geometry.TriangleMesh.create_box(
            gripper_open_length * 0.04,
            gripper_open_length * 0.04,
            gripper_open_length * 0.5,
        )
        gripper_finger.translate(
            np.array(
                [
                    -gripper_open_length * 0.02,
                    -gripper_open_length * 0.02,
                    -gripper_open_length * 0.25,
                ]
            )
        )
        gripper_left = deepcopy(gripper_finger)
        gripper_left = gripper_left.translate(
            np.array(
                [
                    -gripper_open_length * open_ratio * 0.5,
                    0,
                    -gripper_open_length * 0.25,
                ]
            )
        )

        gripper_right = deepcopy(gripper_finger)
        gripper_right = gripper_right.translate(
            np.array(
                [gripper_open_length * open_ratio * 0.5, 0, -gripper_open_length * 0.25]
            )
        )

        gripper_root1 = o3d.geometry.TriangleMesh.create_box(
            gripper_open_length * open_ratio,
            gripper_open_length * 0.04,
            gripper_open_length * 0.04,
        )
        gripper_root1.translate(
            np.array(
                [
                    gripper_open_length * open_ratio * -0.5,
                    gripper_open_length * -0.02,
                    gripper_open_length * -0.02,
                ]
            )
        )
        gripper_root1.translate(
            np.array(
                [
                    0,
                    0,
                    gripper_open_length * -0.5,
                ]
            )
        )

        gripper_root2 = o3d.geometry.TriangleMesh.create_box(
            gripper_open_length * 0.04,
            gripper_open_length * 0.04,
            gripper_open_length * 0.5,
        )
        gripper_root2.translate(
            np.array(
                [
                    gripper_open_length * -0.02,
                    gripper_open_length * -0.02,
                    gripper_open_length * -0.25,
                ]
            )
        )
        gripper_root2.translate(
            np.array(
                [
                    0,
                    0,
                    gripper_open_length * -0.75,
                ]
            )
        )

        gripper_visual = gripper_left + gripper_right + gripper_root1 + gripper_root2
        gripper_visual.compute_vertex_normals()
        return gripper_visual


class Antipodal:
    def __init__(
        self,
        point_a: np.ndarray,
        point_b: np.ndarray,
        normal_a: np.ndarray,
        normal_b: np.ndarray,
    ) -> None:
        """antipodal contact pair

        Args:
            point_a (np.ndarray): position in point a
            point_b (np.ndarray): position in point b
            normal_a (np.ndarray): normal in point a
            normal_b (np.ndarray): normal in point b
        """
        self.point_a = point_a
        self.point_b = point_b
        self.normal_a = normal_a
        self.normal_b = normal_b
        self.dis = np.linalg.norm(point_a - point_b)
        self.angle_cos = self.normal_a.dot(-self.normal_b)
        self.score = self._get_score()
        self._canonical_pose = self._get_canonical_pose()

    def _get_score(self):
        # TODO:  only normal angle is taken into account
        return self.angle_cos

    def get_dis(self, another) -> float:
        """get distance acoording to another antipodal

        Args:
            other (Antipodal): another antipodal

        Returns:
            float: distance
        """
        aa_dis = np.linalg.norm(self.point_a - another.point_a)
        bb_dis = np.linalg.norm(self.point_b - another.point_b)
        ab_dis = np.linalg.norm(self.point_a - another.point_b)
        ba_dis = np.linalg.norm(self.point_b - another.point_a)
        return min(aa_dis, bb_dis, ab_dis, ba_dis)

    def get_dis_arr(self, others) -> np.ndarray:
        """get distance acoording to others antipodals

        Args:
            others (List[Antipodal]): other antipodals

        Returns:
            np.ndarray: distance array
        """
        if not others:
            return np.array([], dtype=float)
        # Vectorized extraction of points using list comprehension and np.array
        other_a = np.array([o.point_a for o in others], dtype=float)
        other_b = np.array([o.point_b for o in others], dtype=float)
        aa_dis = np.linalg.norm(other_a - self.point_a, axis=1)
        ab_dis = np.linalg.norm(other_a - self.point_b, axis=1)
        ba_dis = np.linalg.norm(other_b - self.point_a, axis=1)
        bb_dis = np.linalg.norm(other_b - self.point_b, axis=1)
        dis_arr = np.vstack([aa_dis, ab_dis, ba_dis, bb_dis]).min(axis=0)
        return dis_arr

    def _get_canonical_pose(self) -> np.ndarray:
        """get canonical pose of antipodal contact pair

        Returns:
            np.ndarray: canonical pose
        """
        # assume gripper closing along x_axis
        x_d = self.point_a - self.point_b
        x_d = x_d / np.linalg.norm(x_d)
        y_d = np.cross(np.array([0, 0, 1.0], dtype=float), x_d)
        if np.linalg.norm(y_d) < 1e-4:
            y_d = np.cross(np.array([1, 0, 0.0], dtype=float), x_d)
        y_d = y_d / np.linalg.norm(y_d)
        z_d = np.cross(x_d, y_d)
        pose = np.eye(4, dtype=float)
        pose[:3, 0] = x_d  # rotation x
        pose[:3, 1] = y_d  # rotation y
        pose[:3, 2] = z_d  # rotation z
        pose[:3, 3] = (self.point_a + self.point_b) / 2  # position
        return pose

    def sample_pose(self, sample_num: int = 36) -> np.ndarray:
        """sample parallel gripper grasp poses given antipodal contact pairs

        Args:
            sample_num (int, optional): sample number. Defaults to 36.

        Returns:
            np.ndarray: [sample_num, 4, 4] of float. Sample poses.
        """
        # assume gripper closing along x_axis
        x_d = self._canonical_pose[:3, 0]
        y_d = self._canonical_pose[:3, 1]
        z_d = self._canonical_pose[:3, 2]
        position = self._canonical_pose[:3, 3]
        beta_list = np.linspace(2 * np.pi / sample_num, 2 * np.pi, sample_num)
        grasp_poses = np.empty(shape=(sample_num, 4, 4), dtype=float)
        for i in range(sample_num):
            sample_z = np.sin(beta_list[i]) * y_d + np.cos(beta_list[i]) * z_d
            sample_z = sample_z / np.linalg.norm(sample_z)
            sample_y = np.cross(sample_z, x_d)
            pose = np.eye(4, dtype=float)
            pose[:3, 0] = x_d  # rotation x
            pose[:3, 1] = sample_y  # rotation y
            pose[:3, 2] = sample_z  # rotation z
            pose[:3, 3] = position  # position
            grasp_poses[i] = pose
        return grasp_poses


class AntipodalGenerator:
    def __init__(
        self,
        mesh_o3dt: o3d.t.geometry.TriangleMesh,
        open_length: float,
        min_open_length: float = 0.002,
        max_angle: float = np.pi / 10,
        surface_sample_num: int = 5000,
        layer_num: int = 4,
        sample_each_layer: int = 6,
        nms_ratio: float = 0.02,
        antipodal_sample_num: int = 16,
        unique_id: str = None,
        cache_dir: str = None,
    ):
        """antipodal grasp pose generator

        Args:
            mesh_o3dt (o3d.t.geometry.TriangleMesh): input mesh
            open_length (float): gripper maximum open length
            max_angle (float, optional): maximum grasp direction with surface normal. Defaults to np.pi/10.
            surface_sample_num (int, optional): contact sample number in mesh surface. Defaults to 5000.
            layer_num (int, optional): cone sample layer number . Defaults to 4.
            sample_each_layer (int, optional): cone sample number in each layer. Defaults to 6.
            nms_ratio (float, optional): nms distance ratio. Defaults to 0.02.
            antipodal_sample_num (int, optional): grasp poses sample on each antipodal contact pair. Defaults to 16.
            cache_dir (str, optional): file cache directory. Defaults to None.
        """
        self._antipodal_max_angle = max_angle
        self._open_length = open_length
        self._min_open_length = min_open_length
        self._mesh_o3dt = mesh_o3dt
        verts = mesh_o3dt.vertex.positions.numpy()
        self._center_of_mass = verts.mean(axis=0)
        if unique_id is None:
            unique_file_name = self._get_unique_id(
                mesh_o3dt, open_length, max_angle, surface_sample_num
            )
        else:
            unique_file_name = f"{unique_id}_{str(open_length)}_{str(max_angle)}_{str(surface_sample_num)}"
        if cache_dir is None:
            cache_dir = os.path.join(pathlib.Path.home(), "grasp_pose")
            logger.log_debug(f"Set cache directory to {cache_dir}.")
        if not os.path.isdir(cache_dir):
            os.mkdir(cache_dir)
        cache_file = os.path.join(cache_dir, f"{unique_file_name}.pickle")
        if not os.path.isfile(cache_file):
            # generate cache
            grasp_list = self._generate_cache(
                cache_file,
                mesh_o3dt=mesh_o3dt,
                max_angle=max_angle,
                surface_sample_num=surface_sample_num,
                layer_num=layer_num,
                sample_each_layer=sample_each_layer,
                nms_ratio=nms_ratio,
                antipodal_sample_num=antipodal_sample_num,
            )
        else:
            # load cache
            grasp_list = self._load_cache(cache_file)
        self._grasp_list = grasp_list

    def _get_unique_id(
        self,
        mesh_o3dt: o3d.t.geometry.TriangleMesh,
        open_length: float,
        max_angle: float,
        surface_sample_num: int,
    ) -> str:
        mesh_md5 = get_mesh_md5(mesh_o3dt)
        return (
            f"{mesh_md5}_{str(open_length)}_{str(max_angle)}_{str(surface_sample_num)}"
        )

    def _generate_cache(
        self,
        cache_file: str,
        mesh_o3dt: o3d.t.geometry.TriangleMesh,
        max_angle: float = np.pi / 10,
        surface_sample_num: int = 5000,
        layer_num: int = 4,
        sample_each_layer: int = 6,
        nms_ratio: float = 0.02,
        antipodal_sample_num: int = 16,
    ):
        self._mesh_o3dt = mesh_o3dt
        self._cone_sampler = ConeSampler(
            max_angle=max_angle,
            layer_num=layer_num,
            sample_each_layer=sample_each_layer,
        )
        mesh_o3dt = mesh_o3dt.compute_vertex_normals()
        assert 1e-4 < max_angle < np.pi / 2
        self._mesh_len = self._get_pc_size(mesh_o3dt.vertex.positions.numpy()).max()
        start_time = time.time()
        antipodal_list = self._antipodal_generate(mesh_o3dt, surface_sample_num)
        logger.log_debug(
            f"Antipodal sampling cost {(time.time() - start_time) * 1000} ms."
        )
        logger.log_debug(f"Find {len(antipodal_list)} initial antipodal pairs.")

        valid_antipodal_list = self._antipodal_clean(antipodal_list)

        start_time = time.time()
        nms_antipodal_list = self._antipodal_nms(
            valid_antipodal_list, nms_ratio=nms_ratio
        )
        logger.log_debug(f"NMS cost {(time.time() - start_time) * 1000} ms.")
        logger.log_debug(
            f"There are {len(nms_antipodal_list)} antipodal pair after nms."
        )
        # all poses
        start_time = time.time()
        grasp_poses, score, open_length = self._sample_grasp_pose(
            nms_antipodal_list, antipodal_sample_num
        )
        logger.log_debug(f"Pose sampling cost {(time.time() - start_time) * 1000} ms.")
        logger.log_debug(
            f"There are {grasp_poses.shape[0]} poses after grasp pose sampling."
        )
        # write data
        data_dict = {
            "grasp_poses": grasp_poses,
            "score": score,
            "open_length": open_length,
        }
        pickle.dump(data_dict, open(cache_file, "wb"))
        # TODO: contact pair visualization
        # self.antipodal_visual(nms_antipodal_list)
        grasp_num = grasp_poses.shape[0]
        logger.log_debug(f"Write {grasp_num} poses to pickle file {cache_file}.")
        # Use list comprehension for efficient list construction
        grasp_list = [
            AntipodalGrasp(grasp_poses[i], open_length[i], score[i])
            for i in range(grasp_num)
        ]
        return grasp_list

    def _load_cache(self, cache_file: str):
        data_dict = pickle.load(open(cache_file, "rb"))
        grasp_num = data_dict["grasp_poses"].shape[0]
        logger.log_debug(f"Load {grasp_num} poses from pickle file {cache_file}.")
        # Use list comprehension for efficient list construction
        grasp_list = [
            AntipodalGrasp(
                data_dict["grasp_poses"][i],
                data_dict["open_length"][i],
                data_dict["score"][i],
            )
            for i in range(grasp_num)
        ]
        return grasp_list

    def _get_pc_size(self, vertices):
        return np.array(
            [
                vertices[:, 0].max() - vertices[:, 0].min(),
                vertices[:, 1].max() - vertices[:, 1].min(),
                vertices[:, 2].max() - vertices[:, 2].min(),
            ]
        )

    def _antipodal_generate(
        self, mesh_o3dt: o3d.t.geometry.TriangleMesh, surface_sample_num: int = 5000
    ):
        surface_pcd = mesh_o3dt.to_legacy().sample_points_uniformly(surface_sample_num)
        points = np.array(surface_pcd.points)
        normals = np.array(surface_pcd.normals)
        point_num = points.shape[0]
        scene = o3d.t.geometry.RaycastingScene()
        scene.add_triangles(mesh_o3dt)
        # raycast
        ray_n = self._cone_sampler._ray_num
        ray_num = point_num * ray_n
        ray_begins = np.empty(shape=(ray_num, 3), dtype=float)
        ray_direcs = np.empty(shape=(ray_num, 3), dtype=float)
        origin_normals = np.empty(shape=(ray_num, 3), dtype=float)
        origin_points = np.empty(shape=(ray_num, 3), dtype=float)
        start_time = time.time()
        for i in range(point_num):
            ray_direc = self._cone_sampler.cone_sample_direc(
                normals[i], is_visual=False
            )
            # raycast from outside of object
            ray_begin = points[i] - 2 * self._mesh_len * ray_direc
            ray_direcs[i * ray_n : (i + 1) * ray_n, :] = ray_direc
            ray_begins[i * ray_n : (i + 1) * ray_n, :] = ray_begin
            origin_normals[i * ray_n : (i + 1) * ray_n, :] = normals[i]
            origin_points[i * ray_n : (i + 1) * ray_n, :] = points[i]
        logger.log_debug(f"Cone sampling cost {(time.time() - start_time) * 1000} ms.")

        start_time = time.time()
        rays = o3d.core.Tensor(
            np.hstack([ray_begins, ray_direcs]), dtype=o3d.core.Dtype.Float32
        )
        logger.log_debug(f"Open3d raycast {(time.time() - start_time) * 1000} ms.")

        raycast_rtn = scene.cast_rays(rays)
        hit_dis_all = raycast_rtn["t_hit"].numpy()
        hit_normal_all = raycast_rtn["primitive_normals"].numpy()

        # max_angle_cos = np.cos(self._antipodal_max_angle)
        antipodal_list = []
        # get antipodal points
        start_time = time.time()
        for i in range(ray_num):
            hit_dis = hit_dis_all[i]
            hit_normal = hit_normal_all[i]
            hit_point = ray_begins[i] + ray_direcs[i] * hit_dis
            antipodal_dis = np.linalg.norm(origin_points[i] - hit_point)
            if (
                antipodal_dis > self._min_open_length
                and antipodal_dis < self._open_length
            ):
                # reject thin close object
                antipodal = Antipodal(
                    origin_points[i], hit_point, origin_normals[i], hit_normal
                )
                antipodal_list.append(antipodal)
        logger.log_debug(
            f"Antipodal initialize cost {(time.time() - start_time) * 1000} ms."
        )
        return antipodal_list

    def _sample_grasp_pose(
        self, antipodal_list: List[Antipodal], antipodal_sample_num: int = 36
    ):
        antipodal_num = len(antipodal_list)
        grasp_poses = np.empty(
            shape=(antipodal_sample_num * antipodal_num, 4, 4), dtype=float
        )
        scores = np.empty(shape=(antipodal_sample_num * antipodal_num,), dtype=float)
        open_length = np.empty(
            shape=(antipodal_sample_num * antipodal_num,), dtype=float
        )
        for i in range(antipodal_num):
            grasp_poses[i * antipodal_sample_num : (i + 1) * antipodal_sample_num] = (
                antipodal_list[i].sample_pose(antipodal_sample_num)
            )
            scores[i * antipodal_sample_num : (i + 1) * antipodal_sample_num] = (
                antipodal_list[i].score
            )
            open_length[i * antipodal_sample_num : (i + 1) * antipodal_sample_num] = (
                antipodal_list[i].dis
            )
        return grasp_poses, scores, open_length

    def get_all_grasp(self) -> List[AntipodalGrasp]:
        """get all grasp

        Returns:
            List[AntipodalGrasp]: list of grasp
        """
        return self._grasp_list

    def select_grasp(
        self,
        approach_direction: np.ndarray,
        select_num: int = 20,
        max_angle: float = np.pi / 10,
        select_method: GraspSelectMethod = GraspSelectMethod.NORMAL_SCORE,
    ) -> List[AntipodalGrasp]:
        """Select grasps. Masked by max_angle and sort by grasp score

        Args:
            approach_direction (np.ndarray): gripper approach direction
            select_num (int, optional): select grasp number. Defaults to 10.
            max_angle (float, optional): max angle threshold (angle with surface normal). Defaults to np.pi/10.
            select_method (select_method, optional)
        Return:
            List[AntipodalGrasp]: list of grasp
        """
        grasp_num = len(self._grasp_list)
        all_idx = np.arange(grasp_num)

        # Vectorized extraction of poses and scores using list comprehension
        grasp_poses = np.array([g.pose for g in self._grasp_list], dtype=float)
        scores = np.array([g.score for g in self._grasp_list], dtype=float)
        position = grasp_poses[:, :3, 3]

        # mask acoording to table up direction
        grasp_z = grasp_poses[:, :3, 2]
        direc_dot = (grasp_z * approach_direction).sum(axis=1)
        valid_mask = direc_dot > np.cos(max_angle)
        valid_id = all_idx[valid_mask]

        # sort acoording to different grasp score
        if select_method == GraspSelectMethod.NORMAL_SCORE:
            valid_score = scores[valid_id]
            sort_valid_idx = np.argsort(valid_score)[::-1]  # large is better
        elif select_method == GraspSelectMethod.NEAR_APPROACH:
            position_dot = (position * approach_direction).sum(axis=1)
            valid_height = position_dot[valid_id]
            sort_valid_idx = np.argsort(valid_height)
        elif select_method == GraspSelectMethod.CENTER:
            center_dis = np.linalg.norm(position - self._center_of_mass, axis=1)
            valid_center_dis = center_dis[valid_id]
            sort_valid_idx = np.argsort(valid_center_dis)
        else:
            logger.log_warning(f"select_method {select_method.name} not implemented.")
            # return all grasp
            return self._grasp_list

        # get best score sample index
        result_num = min(len(sort_valid_idx), select_num)
        best_valid_idx = sort_valid_idx[:result_num]
        best_idx = valid_id[best_valid_idx]

        # Use list comprehension for faster list construction
        return [self._grasp_list[idx] for idx in best_idx]

    def _antipodal_nms(
        self, antipodal_list: List[Antipodal], nms_ratio: float = 0.02
    ) -> List[Antipodal]:
        antipodal_num = len(antipodal_list)
        nms_mask = np.empty(shape=(antipodal_num,), dtype=bool)
        nms_mask.fill(True)
        score_list = np.empty(shape=(antipodal_num,), dtype=float)

        for i in range(antipodal_num):
            score_list[i] = antipodal_list[i].score

        sort_idx = np.argsort(score_list)[::-1]

        dis_th = self._mesh_len * nms_ratio
        for i in range(antipodal_num):
            if not nms_mask[sort_idx[i]]:
                continue
            antipodal_max = antipodal_list[sort_idx[i]]
            other_antipodal = []
            other_idx = []
            for j in range(i + 1, antipodal_num):
                if nms_mask[sort_idx[j]]:
                    other_antipodal.append(antipodal_list[sort_idx[j]])
                    other_idx.append(sort_idx[j])
            dis_arr = antipodal_max.get_dis_arr(other_antipodal)
            invalid_mask = dis_arr < dis_th
            for j, flag in enumerate(invalid_mask):
                if flag:
                    nms_mask[other_idx[j]] = False
        nms_antipodal_list = []
        for i in range(antipodal_num):
            if nms_mask[sort_idx[i]]:
                nms_antipodal_list.append(antipodal_list[sort_idx[i]])

        # TODO: nms validation check. remove in future
        # antipodal_num = len(nms_antipodal_list)
        # for i in range(antipodal_num):
        #     for j in range(i + 1, antipodal_num):
        #         antipodal_dis = nms_antipodal_list[i].get_dis(nms_antipodal_list[j])
        #         if antipodal_dis < dis_th:
        #             logger.log_warning(f"find near antipodal {i} and {j} with dis {antipodal_dis}")
        return nms_antipodal_list

    def _antipodal_clean(self, antipodal_list: List[Antipodal]):
        # TODO: need collision checker

        valid_antipodal = []
        max_angle_cos = np.cos(self._antipodal_max_angle)
        for antipodal in antipodal_list:
            if (
                1e-5 < antipodal.dis < self._open_length
                and antipodal.angle_cos > max_angle_cos
            ):
                valid_antipodal.append(antipodal)
        return valid_antipodal

    def antipodal_visual(self, antipodal_list):
        mesh_visual = self._mesh_o3dt.to_legacy()
        antipodal_num = len(antipodal_list)
        draw_points = np.empty(shape=(antipodal_num * 2, 3), dtype=float)
        draw_lines = np.empty(shape=(antipodal_num, 2), dtype=int)
        for i in range(antipodal_num):
            direc = antipodal_list[i].point_b - antipodal_list[i].point_a
            direc = direc / np.linalg.norm(direc)
            anti_begin = antipodal_list[i].point_a - direc * 0.005
            anti_end = antipodal_list[i].point_b + direc * 0.005
            draw_points[i * 2] = anti_begin
            draw_points[i * 2 + 1] = anti_end
            draw_lines[i] = np.array([i * 2, i * 2 + 1], dtype=int)

        line_set = o3d.geometry.LineSet(
            points=o3d.utility.Vector3dVector(draw_points),
            lines=o3d.utility.Vector2iVector(draw_lines),
        )
        draw_colors = np.empty(shape=(antipodal_num, 3), dtype=float)
        draw_colors[:] = np.array([0.0, 1.0, 1.0])
        line_set.colors = o3d.utility.Vector3dVector(draw_colors)
        o3d.visualization.draw_geometries([line_set, mesh_visual])

    def grasp_pose_visual(
        self, grasp_list: List[AntipodalGrasp]
    ) -> List[o3d.t.geometry.TriangleMesh]:
        """visualize grasp pose

        Args:
            grasp_list (List[AntipodalGrasp]): list of grasp

        Returns:
            List[o3d.t.geometry.TriangleMesh]: list of visualization mesh
        """
        pose_num = len(grasp_list)
        visual_mesh_list = [self._mesh_o3dt.compute_vertex_normals()]

        max_angle_cos = np.cos(self._antipodal_max_angle)

        for i in range(pose_num):
            grasp_mesh = grasp_list[i].grasp_pose_visual_mesh(
                gripper_open_length=self._open_length
            )
            grasp_mesh.transform(grasp_list[i].pose)
            # low score: red | hight score: blue
            score_ratio = (grasp_list[i].score - max_angle_cos) / (1 - max_angle_cos)
            score_ratio = min(1.0, score_ratio)
            score_ratio = max(0.0, score_ratio)
            grasp_mesh.paint_uniform_color(np.array([1 - score_ratio, 0, score_ratio]))
            visual_mesh_list.append(o3d.t.geometry.TriangleMesh.from_legacy(grasp_mesh))
        return visual_mesh_list



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from __future__ import annotations

import copy
import torch
import dexsim
import numpy as np

from typing import Dict, Union
from functools import cached_property

from dexsim.engine import MaterialInst, Material
from embodichain.lab.sim.utility import is_rt_enabled
from embodichain.utils import configclass, logger


@configclass
class VisualMaterialCfg:
    """Configuration for visual material with PBR properties for rasterization and ray tracing."""

    uid: str = "default_mat"

    # Basic PBR properties
    base_color: list = [0.5, 0.5, 0.5, 1.0]
    """Base color/diffuse color (RGBA)"""

    metallic: float = 0.0
    """Metallic factor (0.0 = dielectric, 1.0 = metallic)"""

    roughness: float = 0.5
    """Surface roughness (0.0 = smooth, 1.0 = rough)"""

    # Additional PBR properties
    emissive: list = [0.0, 0.0, 0.0]  # Emissive color (RGB)
    emissive_intensity: float = 1.0  # Emissive intensity multiplier

    # Texture maps
    base_color_texture: str = None
    """Base color texture map"""

    metallic_texture: str = None
    """Metallic map"""

    roughness_texture: str = None
    """Roughness map"""

    normal_texture: str = None
    """Normal map"""

    ao_texture: str = None
    """Ambient occlusion map"""

    # Ray tracing specific properties
    ior: float = 1.5
    """Index of refraction for PBR materials, only used in ray tracing."""

    material_type: str = "BRDF"
    """Ray tracing material type. Options: 'BRDF', 'BTDF', 'BSDF'"""

    # Currently disabled properties
    # subsurface: float = 0.0  # Subsurface scattering factor
    # subsurface_color: list = [1.0, 1.0, 1.0]  # Subsurface scattering color

    @classmethod
    def from_dict(cls, cfg_dict: dict) -> VisualMaterialCfg:
        base = cls()
        for k, v in cfg_dict.items():
            if hasattr(base, k):
                setattr(base, k, v)
            else:
                logger.log_warning(f"Unknown field '{k}' in VisualMaterialCfg.")
        return base


class VisualMaterial:
    """Visual material definition in the simulation environment.

    A visual material is actually a material template from which material instances can be created.
    It holds multiple material instances, which is used to assign to different objects in the environment.
    """

    RT_MATERIAL_TYPES = [
        "BRDF",
        "BTDF",
        "BSDF",
    ]

    MAT_TYPE_MAPPING: Dict[str, str] = {
        "BRDF": "BRDF_GGX_SMITH",
        "BTDF": "BTDF_GGX_SMITH",
        "BSDF": "BSDF_GGX_SMITH",
    }

    def __init__(self, cfg: VisualMaterialCfg, mat: Material):
        if cfg.material_type not in self.RT_MATERIAL_TYPES:
            logger.log_error(
                f"Invalid material_type '{cfg.material_type}'. "
                f"Supported types: {self.RT_MATERIAL_TYPES}"
            )

        self.uid = cfg.uid
        self.cfg = copy.deepcopy(cfg)
        self._mat = mat
        self._mat_inst_list: list[str] = []

        self._default_mat_inst = self.create_instance(self.uid)

    @cached_property
    def is_rt_enabled(self) -> bool:
        return is_rt_enabled()

    @property
    def mat(self) -> Material:
        return self._mat

    @property
    def inst(self) -> VisualMaterialInst:
        return self._default_mat_inst

    def set_default_properties(
        self, mat_inst: VisualMaterialInst, cfg: VisualMaterialCfg
    ) -> None:
        mat_inst.set_base_color(cfg.base_color)
        mat_inst.set_metallic(cfg.metallic)
        mat_inst.set_roughness(cfg.roughness)
        mat_inst.set_emissive(cfg.emissive)
        # mat_inst.set_emissive_intensity(self.cfg.emissive_intensity)  # Unimplemented

        mat_inst.set_base_color_texture(cfg.base_color_texture)
        mat_inst.set_metallic_texture(cfg.metallic_texture)
        mat_inst.set_roughness_texture(cfg.roughness_texture)
        mat_inst.set_normal_texture(cfg.normal_texture)
        mat_inst.set_ao_texture(cfg.ao_texture)

        if self.is_rt_enabled:
            mat_inst.set_ior(cfg.ior)
            mat_inst.mat.update_pbr_material_type(
                self.MAT_TYPE_MAPPING[cfg.material_type]
            )

    def create_instance(self, uid: str) -> VisualMaterialInst:
        """Create a new material instance from this material template.

        Note:
            - If the uid already exists, the existing instance will be returned.

        Args:
            uid (str): Unique identifier for the material instance.

        Returns:
            VisualMaterialInst: The created material instance.
        """
        inst = VisualMaterialInst(uid, self._mat)
        # TODO: Support change default properties for material.
        # This will improve the instance creation efficiency.
        self.set_default_properties(inst, self.cfg)
        self._mat_inst_list.append(uid)
        return inst

    def get_default_instance(self) -> VisualMaterialInst:
        """Get the default material instance created with the same uid as the material template.

        Returns:
            VisualMaterialInst: The default material instance.
        """
        return self._default_mat_inst

    def get_instance(self, uid: str) -> VisualMaterialInst:
        """Get an existing material instance by its uid.

        Args:
            uid (str): Unique identifier for the material instance.

        Returns:
            VisualMaterialInst: The material instance.
        """
        if uid not in self._mat_inst_list:
            logger.log_error(f"Material instance with uid '{uid}' does not exist.")
        return VisualMaterialInst(uid, self._mat)


class VisualMaterialInst:
    """Instance of a visual material in the simulation environment."""

    def __init__(self, uid: str, mat: Material):
        self.uid = uid
        self._mat = mat

        # Init properties with default values
        self.base_color = [0.5, 0.5, 0.5, 1.0]
        self.metallic = 0.0
        self.roughness = 0.5
        self.emissive = [0.0, 0.0, 0.0]
        self.emissive_intensity = 1.0
        self.base_color_texture = None
        self.metallic_texture = None
        self.roughness_texture = None
        self.normal_texture = None
        self.ao_texture = None
        self.ior = 1.5
        # self.subsurface = 0.0

    @property
    def mat(self) -> MaterialInst:
        return self._mat.get_inst(self.uid)

    def set_base_color(self, color: list) -> None:
        """Set base color/diffuse color."""
        self.base_color = color
        self.mat.set_base_color(color)

    def set_metallic(self, metallic: float) -> None:
        """Set metallic factor."""
        self.metallic = metallic
        inst = self._mat.get_inst(self.uid)
        inst.set_metallic(metallic)

    def set_roughness(self, roughness: float) -> None:
        """Set surface roughness."""
        self.roughness = roughness
        inst = self._mat.get_inst(self.uid)
        inst.set_roughness(roughness)

    def set_emissive(self, emissive: list) -> None:
        """Set emissive color."""
        self.emissive = emissive
        value = np.zeros(4)
        value[0:3] = emissive
        inst = self._mat.get_inst(self.uid)
        inst.set_emissive(value)

    def set_emissive_intensity(self, intensity: float) -> None:
        """Set emissive intensity multiplier."""
        logger.log_error("Unimplemented: set_emissive_intensity")

    def set_base_color_texture(
        self, texture_path: str = None, texture_data: torch.Tensor | None = None
    ) -> None:
        """Set base color texture from file path or texture data.

        Args:
            texture_path: Path to texture file
            texture_data: Texture data as a torch.Tensor
        """
        if texture_path is not None and texture_data is not None:
            logger.log_warning(
                "Both texture_path and texture_data are provided. Using texture_path."
            )

        if texture_path is not None:
            self.base_color_texture = texture_path
            inst = self._mat.get_inst(self.uid)
            inst.set_base_color_map(texture_path)
        elif texture_data is not None:
            self.base_color_texture = texture_data
            inst = self._mat.get_inst(self.uid)

            # TODO: Optimize texture creation method.
            world = dexsim.default_world()
            env = world.get_env()
            color_texture = env.create_color_texture(
                texture_data.cpu().numpy(), has_alpha=True
            )
            inst.set_base_color_map(color_texture)

    def set_metallic_texture(
        self, texture_path: str = None, texture_data: torch.Tensor | None = None
    ) -> None:
        """Set metallic texture from file path or texture data.

        Args:
            texture_path: Path to texture file
            texture_data: Texture data as a torch.Tensor
        """
        if texture_path is not None and texture_data is not None:
            logger.log_warning(
                "Both texture_path and texture_data are provided. Using texture_path."
            )

        if texture_path is not None:
            self.metallic_texture = texture_path
            inst = self._mat.get_inst(self.uid)
            inst.set_metallic_map(texture_path)
        elif texture_data is not None:
            self.metallic_texture = texture_data
            inst = self._mat.get_inst(self.uid)

            # TODO: Optimize texture creation method.
            world = dexsim.default_world()
            env = world.get_env()
            metallic_texture = env.create_color_texture(
                texture_data.cpu().numpy(), has_alpha=False
            )
            inst.set_metallic_map(metallic_texture)

    def set_roughness_texture(
        self, texture_path: str = None, texture_data: torch.Tensor | None = None
    ) -> None:
        """Set roughness texture from file path or texture data.

        Args:
            texture_path: Path to texture file
            texture_data: Texture data as a torch.Tensor
        """
        if texture_path is not None and texture_data is not None:
            logger.log_warning(
                "Both texture_path and texture_data are provided. Using texture_path."
            )

        if texture_path is not None:
            self.roughness_texture = texture_path
            inst = self._mat.get_inst(self.uid)
            inst.set_roughness_map(texture_path)
        elif texture_data is not None:
            self.roughness_texture = texture_data
            inst = self._mat.get_inst(self.uid)

            # TODO: Optimize texture creation method.
            world = dexsim.default_world()
            env = world.get_env()
            roughness_texture = env.create_color_texture(
                texture_data.cpu().numpy(), has_alpha=False
            )
            inst.set_roughness_map(roughness_texture)

    def set_normal_texture(
        self, texture_path: str = None, texture_data: torch.Tensor | None = None
    ) -> None:
        """Set normal texture from file path or texture data.

        Args:
            texture_path: Path to texture file
            texture_data: Texture data as a torch.Tensor
        """
        if texture_path is not None and texture_data is not None:
            logger.log_warning(
                "Both texture_path and texture_data are provided. Using texture_path."
            )

        if texture_path is not None:
            self.normal_texture = texture_path
            inst = self._mat.get_inst(self.uid)
            inst.set_normal_map(texture_path)
        elif texture_data is not None:
            self.normal_texture = texture_data
            inst = self._mat.get_inst(self.uid)

            # TODO: Optimize texture creation method.
            world = dexsim.default_world()
            env = world.get_env()
            normal_texture = env.create_color_texture(
                texture_data.cpu().numpy(), has_alpha=False
            )
            inst.set_normal_map(normal_texture)

    def set_ao_texture(
        self, texture_path: str = None, texture_data: torch.Tensor | None = None
    ) -> None:
        """Set ambient occlusion texture from file path or texture data.

        Args:
            texture_path: Path to texture file
            texture_data: Texture data as a torch.Tensor
        """
        if texture_path is not None and texture_data is not None:
            logger.log_warning(
                "Both texture_path and texture_data are provided. Using texture_path."
            )

        if texture_path is not None:
            self.ao_texture = texture_path
            inst = self._mat.get_inst(self.uid)
            inst.set_ao_map(texture_path)
        elif texture_data is not None:
            self.ao_texture = texture_data
            inst = self._mat.get_inst(self.uid)

            # TODO: Optimize texture creation method.
            world = dexsim.default_world()
            env = world.get_env()
            ao_texture = env.create_color_texture(
                texture_data.cpu().numpy(), has_alpha=False
            )
            inst.set_ao_map(ao_texture)

    def set_ior(self, ior: float) -> None:
        """Set index of refraction."""
        if is_rt_enabled() is False:
            logger.log_debug("Ray Tracing rendering not enabled, ignoring IOR setting.")
            return
        self.ior = ior
        inst = self._mat.get_inst(self.uid)
        inst.set_rt_param("ior", ior)



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from __future__ import annotations

import json
import sys
import torch

from copy import deepcopy
from functools import partial
from typing import TYPE_CHECKING, Dict, Type

import gymnasium as gym
from gymnasium.envs.registration import EnvSpec as GymEnvSpec
from gymnasium.envs.registration import WrapperSpec

from dexsim.utility import log_warning

if TYPE_CHECKING:
    from embodichain.lab.gym.envs import BaseEnv


class EnvSpec:
    def __init__(
        self,
        uid: str,
        cls: Type[BaseEnv],
        max_episode_steps=None,
        default_kwargs: dict = None,
    ):
        """A specification for a Embodied environment."""
        self.uid = uid
        self.cls = cls
        self.max_episode_steps = max_episode_steps
        self.default_kwargs = {} if default_kwargs is None else default_kwargs

    def make(self, **kwargs):
        _kwargs = self.default_kwargs.copy()
        _kwargs.update(kwargs)
        return self.cls(**_kwargs)

    @property
    def gym_spec(self):
        """Return a gym EnvSpec for this env"""
        entry_point = self.cls.__module__ + ":" + self.cls.__name__
        return GymEnvSpec(
            self.uid,
            entry_point,
            max_episode_steps=self.max_episode_steps,
            kwargs=self.default_kwargs,
        )


REGISTERED_ENVS: Dict[str, EnvSpec] = {}


def register(
    name: str, cls: Type[BaseEnv], max_episode_steps=None, default_kwargs: dict = None
):
    """Register a Embodied environment."""

    # hacky way to avoid circular import errors when users inherit a task in DexSim and try to register it themselves
    from embodichain.lab.gym.envs import BaseEnv, BaseEnv

    if name in REGISTERED_ENVS:
        log_warning(f"Env {name} already registered")
    if not (issubclass(cls, BaseEnv) or issubclass(cls, BaseEnv)):
        raise TypeError(f"Env {name} must inherit from BaseEnv or BaseEnv")
    REGISTERED_ENVS[name] = EnvSpec(
        name, cls, max_episode_steps=max_episode_steps, default_kwargs=default_kwargs
    )


class TimeLimitWrapper(gym.Wrapper):
    """like the standard gymnasium timelimit wrapper but fixes truncated variable to be a batched array"""

    def __init__(self, env: gym.Env, max_episode_steps: int):
        super().__init__(env)
        prev_frame_locals = sys._getframe(1).f_locals
        frame = sys._getframe(1)
        # check for user supplied max_episode_steps during gym.make calls
        if frame.f_code.co_name == "make" and "max_episode_steps" in prev_frame_locals:
            if prev_frame_locals["max_episode_steps"] is not None:
                max_episode_steps = prev_frame_locals["max_episode_steps"]
            # do some wrapper surgery to remove the previous timelimit wrapper
            # with gymnasium 0.29.1, this will remove the timelimit wrapper and nothing else.
            curr_env = env
            while curr_env is not None:
                if isinstance(curr_env, gym.wrappers.TimeLimit):
                    self.env = curr_env.env
                    break
        self._max_episode_steps = self.base_env.max_episode_steps

    @property
    def base_env(self) -> BaseEnv:
        return self.env.unwrapped

    @property
    def device(self) -> torch.device:
        return self.base_env.device

    @property
    def num_envs(self) -> int:
        return self.base_env.num_envs

    def step(self, action):
        observation, reward, terminated, truncated, info = self.env.step(action)
        truncated = truncated | (self.base_env.elapsed_steps >= self._max_episode_steps)
        return observation, reward, terminated, truncated, info


def make(env_id, **kwargs):
    """Instantiate a Embodied environment.

    Args:
        env_id (str): Environment ID.
        as_gym (bool, optional): Add TimeLimit wrapper as gym.
        **kwargs: Keyword arguments to pass to the environment.
    """
    if env_id not in REGISTERED_ENVS:
        raise KeyError("Env {} not found in registry".format(env_id))
    env_spec = REGISTERED_ENVS[env_id]

    env = env_spec.make(**kwargs)
    return env


def make_vec(env_id, **kwargs):
    env = gym.make(env_id, **kwargs)
    return env


def register_env(uid: str, max_episode_steps=None, override=False, **kwargs):
    """A decorator to register Embodied environments.

    Args:
        uid (str): unique id of the environment.
        max_episode_steps (int): maximum number of steps in an episode.
        override (bool): whether to override the environment if it is already registered.

    Notes:
        - `max_episode_steps` is processed differently from other keyword arguments in gym.
          `gym.make` wraps the env with `gym.wrappers.TimeLimit` to limit the maximum number of steps.
        - `gym.EnvSpec` uses kwargs instead of **kwargs!
    """
    try:
        json.dumps(kwargs)
    except TypeError:
        raise RuntimeError(
            f"You cannot register_env with non json dumpable kwargs, e.g. classes or types. If you really need to do this, it is recommended to create a mapping of string to the unjsonable data and to pass the string in the kwarg and during env creation find the data you need"
        )

    def _register_env(cls):
        cls = register_env_function(cls, uid, override, max_episode_steps, **kwargs)
        return cls

    return _register_env


def register_env_function(cls, uid, override=False, max_episode_steps=None, **kwargs):
    if uid in REGISTERED_ENVS:
        if override:
            from gymnasium.envs.registration import registry

            log_warning(f"Override registered env {uid}")
            REGISTERED_ENVS.pop(uid)
            registry.pop(uid)
        else:
            log_warning(f"Env {uid} is already registered. Skip registration.")
            return cls

    register(
        uid,
        cls,
        max_episode_steps=max_episode_steps,
        default_kwargs=deepcopy(kwargs),
    )

    # Register for gym
    gym.register(
        uid,
        entry_point=partial(make, env_id=uid),
        vector_entry_point=partial(make_vec, env_id=uid),
        max_episode_steps=max_episode_steps,
        disable_env_checker=True,  # Temporary solution as we allow empty observation spaces
        kwargs=deepcopy(kwargs),
    )
    return cls



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import pytest
import torch

from embodichain.lab.sim import (
    SimulationManager,
    SimulationManagerCfg,
)
from embodichain.lab.sim.objects import Articulation, RigidObject
from embodichain.lab.sim.cfg import (
    ArticulationCfg,
    RigidObjectCfg,
    JointDrivePropertiesCfg,
    RigidBodyAttributesCfg,
)
from embodichain.lab.sim.shapes import MeshCfg
from embodichain.data import get_data_path

NUM_ARENAS = 1


class BaseUsdTest:
    """Shared test logic for CPU and CUDA."""

    def setup_simulation(self, sim_device):
        config = SimulationManagerCfg(
            headless=True, sim_device=sim_device, num_envs=NUM_ARENAS, enable_rt=False
        )
        self.sim = SimulationManager(config)

        if sim_device == "cuda" and getattr(self.sim, "is_use_gpu_physics", False):
            self.sim.init_gpu_physics()

    def test_import_rigid(self):
        default_attr = RigidBodyAttributesCfg()
        sugar_box_path = get_data_path("SugarBox/sugar_box_usd/sugar_box.usda")
        sugar_box: RigidObject = self.sim.add_rigid_object(
            cfg=RigidObjectCfg(
                uid="sugar_box",
                shape=MeshCfg(fpath=sugar_box_path),
                body_type="dynamic",
                use_usd_properties=False,
                init_pos=[0.0, 1.0, 0.1],
                attrs=default_attr,
            )
        )
        body0 = sugar_box._entities[0].get_physical_body()
        print(sugar_box._entities[0].get_physical_attr())
        assert pytest.approx(body0.get_mass()) == default_attr.mass
        assert pytest.approx(body0.get_linear_damping()) == default_attr.linear_damping
        assert (
            pytest.approx(body0.get_angular_damping()) == default_attr.angular_damping
        )
        assert body0.get_solver_iteration_counts() == (
            default_attr.min_position_iters,
            default_attr.min_velocity_iters,
        )

    def test_import_articulation(self):
        default_drive = JointDrivePropertiesCfg()
        h1_path = get_data_path("UnitreeH1Usd/H1_usd/h1.usd")
        h1: Articulation = self.sim.add_articulation(
            cfg=ArticulationCfg(
                uid="h1",
                fpath=h1_path,
                build_pk_chain=False,
                use_usd_properties=False,
                init_pos=[0.0, 0.0, 1.2],
                drive_pros=default_drive,
            )
        )

        stiffness = h1.body_data.joint_stiffness
        damping = h1.body_data.joint_damping
        max_force = h1.body_data.qf_limits
        print(f"All joint stiffness: {stiffness}")
        print(f"All joint damping: {damping}")
        print(f"All joint max force: {max_force}")
        expected_stiffness = default_drive.stiffness
        assert torch.allclose(
            stiffness, torch.tensor(expected_stiffness, dtype=torch.float32)
        )
        expectied_damping = default_drive.damping
        assert torch.allclose(
            damping, torch.tensor(expectied_damping, dtype=torch.float32)
        )

    def test_usd_properties(self):
        """In this test, we set use_usd_properties=True to verify that the USD properties are correctly applied."""
        h1_path = get_data_path("UnitreeH1Usd/H1_usd/h1.usd")
        h1: Articulation = self.sim.add_articulation(
            cfg=ArticulationCfg(
                uid="h1_beta",
                fpath=h1_path,
                build_pk_chain=False,
                use_usd_properties=True,
                init_pos=[1.0, 0.0, 1.2],
            )
        )

        stiffness = h1.body_data.joint_stiffness
        damping = h1.body_data.joint_damping
        max_force = h1.body_data.qf_limits
        print(f"All joint stiffness: {stiffness}")
        print(f"All joint damping: {damping}")
        print(f"All joint max force: {max_force}")
        expected_stiffness = 10000000.0
        assert torch.allclose(
            stiffness, torch.tensor(expected_stiffness, dtype=torch.float32)
        )

        joint_names = h1.joint_names
        print(f"Joint names: {joint_names}")

        target_joint_name = "left_hip_yaw_joint"
        if target_joint_name in joint_names:
            joint_idx = joint_names.index(target_joint_name)
            # check for the first instance
            assert torch.isclose(
                stiffness[0, joint_idx], torch.tensor(10000000.0, dtype=torch.float32)
            )
            assert torch.isclose(
                damping[0, joint_idx], torch.tensor(0.0, dtype=torch.float32)
            )
            assert torch.isclose(
                max_force[0, joint_idx], torch.tensor(200.0, dtype=torch.float32)
            )

        sugar_box_path = get_data_path("SugarBox/sugar_box_usd/sugar_box.usda")
        sugar_box: RigidObject = self.sim.add_rigid_object(
            cfg=RigidObjectCfg(
                uid="sugar_box_beta",
                shape=MeshCfg(fpath=sugar_box_path),
                body_type="dynamic",
                use_usd_properties=True,
                init_pos=[1.0, 1.0, 0.1],
            )
        )
        body0 = sugar_box._entities[0].get_physical_body()
        print(sugar_box._entities[0].get_physical_attr())
        assert pytest.approx(body0.get_mass(), 0.001) == 0.514
        # TODO: nvidia physx attrs in usd currently are not fully suported
        # assert(body0.get_linear_damping()==0)
        # assert(body0.get_angular_damping()==0.05)
        # assert(body0.get_solver_iteration_counts()==(4, 1))
        # assert(body0.get_max_angular_velocity()==100)

    def export_usd(self):
        self.sim.export_usd("test_export.usda")

    def teardown_method(self):
        """Clean up resources after each test method."""
        self.sim.destroy()


class TestUsdCPU(BaseUsdTest):
    def setup_method(self):
        self.setup_simulation("cpu")


@pytest.mark.skip(reason="Skipping CUDA tests temporarily")
class TestUsdCUDA(BaseUsdTest):
    def setup_method(self):
        self.setup_simulation("cuda")


if __name__ == "__main__":
    test = TestUsdCPU()
    test.setup_method()
    test.test_import_rigid()
    test.test_import_articulation()
    test.export_usd()
    test.test_usd_properties()

    # from IPython import embed
    # embed()



# ----------------------------------------------------------------------------
# Copyright (c) 2021-2026 DexForce Technology Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from __future__ import annotations

from typing import List, Dict, Union, TYPE_CHECKING, Any
from dataclasses import MISSING
from embodichain.utils import configclass, is_configclass, logger

if TYPE_CHECKING:
    from embodichain.lab.sim.material import VisualMaterialCfg


@configclass
class LoadOption:

    rebuild_normals: bool = False
    """Whether to rebuild normals for the shape. Defaults to False."""

    rebuild_tangent: bool = False
    """Whether to rebuild tangents for the shape. Defaults to False."""

    rebuild_3rdnormal: bool = False
    """Whether to rebuild the normal for the shape using 3rd party library. Defaults to False."""

    rebuild_3rdtangent: bool = False
    """Whether to rebuild the tangent for the shape using 3rd party library. Defaults to False."""

    smooth: float = -1.0
    """Angle threshold (in degrees) for smoothing normals. Defaults to -1.0 (no smoothing)."""

    @classmethod
    def from_dict(cls, init_dict: Dict[str, Any]) -> LoadOption:
        """Initialize the configuration from a dictionary."""
        cfg = cls()
        for key, value in init_dict.items():
            if hasattr(cfg, key):
                setattr(cfg, key, value)
            else:
                logger.log_warning(
                    f"Key '{key}' not found in {cfg.__class__.__name__}."
                )
        return cfg


@configclass
class ShapeCfg:

    shape_type: str = MISSING
    """Type of the shape. Must be specified in subclasses."""

    visual_material: VisualMaterialCfg | None = None
    """Configuration parameters for the visual material of the shape. Defaults to None."""

    @classmethod
    def from_dict(cls, init_dict: Dict[str, Any]) -> ShapeCfg:
        """Initialize the configuration from a dictionary."""
        from embodichain.utils.utility import get_class_instance

        if "shape_type" not in init_dict:
            logger.log_error("shape type must be specified in the configuration.")

        cfg = get_class_instance(
            "embodichain.lab.sim.shapes", init_dict["shape_type"] + "Cfg"
        )()
        for key, value in init_dict.items():
            if hasattr(cfg, key):
                attr = getattr(cfg, key)
                if key == "visual_material" and isinstance(value, dict):
                    setattr(
                        cfg,
                        key,
                        VisualMaterialCfg.from_dict(value),
                    )
                elif is_configclass(attr):
                    setattr(cfg, key, attr.from_dict(value))
                else:
                    setattr(cfg, key, value)
            else:
                logger.log_warning(
                    f"Key '{key}' not found in {cfg.__class__.__name__}."
                )
        return cfg


@configclass
class MeshCfg(ShapeCfg):
    """Configuration parameters for a triangle mesh shape."""

    shape_type: str = "Mesh"

    fpath: str = MISSING
    """File path to the shape mesh file."""

    load_option: LoadOption = LoadOption()
    """Options for loading and processing the shape."""

    compute_uv: bool = False
    """Whether to compute UV coordinates for the shape. Defaults to False.
    
    If the shape already has UV coordinates, setting this to True will recompute and overwrite them.
    """

    project_direction: List[float] = [1.0, 1.0, 1.0]
    """Direction to project the UV coordinates. Defaults to [1.0, 1.0, 1.0]."""


@configclass
class CubeCfg(ShapeCfg):
    """Configuration parameters for a cube shape."""

    shape_type: str = "Cube"

    size: List[float] = [1.0, 1.0, 1.0]
    """Size of the cube (in m) as [length, width, height]."""


@configclass
class SphereCfg(ShapeCfg):
    """Configuration parameters for a sphere shape."""

    shape_type: str = "Sphere"

    radius: float = 1.0
    """Radius of the sphere (in m)."""

    resolution: int = 20
    """Resolution of the sphere mesh. Defaults to 20."""
